# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from datetime import datetime
from dateutil import tz
from jsonschema.exceptions import ValidationError

from c7n.filters import FilterValidationError
from c7n.resources import ec2
from c7n.resources.ec2 import actions, QueryFilter
from c7n import tags, utils

from common import BaseTest


class TestTagAugmentation(BaseTest):

    def test_tag_augment_empty(self):
        session_factory = self.replay_flight_data(
            'test_ec2_augment_tag_empty')
        # recording was modified to be sans tags
        ec2 = session_factory().client('ec2')
        policy = self.load_policy({
            'name': 'ec2-tags',
            'resource': 'ec2'},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 0)

    def test_tag_augment(self):
        session_factory = self.replay_flight_data(
            'test_ec2_augment_tags')
        # recording was modified to be sans tags
        ec2 = session_factory().client('ec2')
        policy = self.load_policy({
            'name': 'ec2-tags',
            'resource': 'ec2',
            'filters': [
                {'tag:Env': 'Production'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestMetricFilter(BaseTest):

    def test_metric_filter(self):
        session_factory = self.replay_flight_data(
            'test_ec2_metric')
        ec2 = session_factory().client('ec2')
        policy = self.load_policy({
            'name': 'ec2-utilization',
            'resource': 'ec2',
            'filters': [
                {'type': 'metrics',
                 'name': 'CPUUtilization',
                 'days': 3,
                 'value': 1.5}
            ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestHealthEventsFilter(BaseTest):
    def test_ec2_health_events_filter(self):
        session_factory = self.replay_flight_data(
            'test_ec2_health_events_filter')
        policy = self.load_policy({
            'name': 'ec2-health-events-filter',
            'resource': 'ec2',
            'filters': [
                {'type': 'health-event'}
            ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestTagTrim(BaseTest):

    def test_ec2_tag_trim(self):
        self.patch(tags.TagTrim, 'max_tag_count', 10)
        session_factory = self.replay_flight_data(
            'test_ec2_tag_trim')
        ec2 = session_factory().client('ec2')
        start_tags = {
            t['Key']: t['Value'] for t in
            ec2.describe_tags(
                Filters=[{'Name': 'resource-id',
                          'Values': ['i-fdb01920']}])['Tags']}
        policy = self.load_policy({
            'name': 'ec2-tag-trim',
            'resource': 'ec2',
            'filters': [
                {'type': 'tag-count', 'count': 10}],
            'actions': [
                {'type': 'tag-trim',
                 'space': 1,
                 'preserve': [
                     'Name',
                     'Env',
                     'Account',
                     'Platform',
                     'Classification',
                     'Planet'
                     ]}
                ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        end_tags = {
            t['Key']: t['Value'] for t in
            ec2.describe_tags(
                Filters=[{'Name': 'resource-id',
                          'Values': ['i-fdb01920']}])['Tags']}

        self.assertEqual(len(start_tags)-1, len(end_tags))
        self.assertTrue('Containers' in start_tags)
        self.assertFalse('Containers' in end_tags)


class TestVolumeFilter(BaseTest):

    def test_ec2_attached_ebs_filter(self):
        session_factory = self.replay_flight_data(
            'test_ec2_attached_ebs_filter')
        policy = self.load_policy({
            'name': 'ec2-unencrypted-vol',
            'resource': 'ec2',
            'filters': [
                {'type': 'ebs',
                 'key': 'Encrypted',
                 'value': False}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    # DISABLED / Re-record flight data on public account
    def test_ec2_attached_volume_skip_block(self):
        session_factory = self.replay_flight_data(
            'test_ec2_attached_ebs_filter')
        policy = self.load_policy({
            'name': 'ec2-unencrypted-vol',
            'resource': 'ec2',
            'filters': [
                {'type': 'ebs',
                 'skip-devices': ['/dev/sda1', '/dev/xvda', '/dev/sdb1'],
                 'key': 'Encrypted',
                 'value': False}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 0)


class TestResizeInstance(BaseTest):

    def test_ec2_resize(self):
        # preconditions - three instances (2 m4.4xlarge, 1 m4.1xlarge)
        # one of the instances stopped
        session_factory = self.replay_flight_data('test_ec2_resize')
        policy = self.load_policy({
            'name': 'ec2-resize',
            'resource': 'ec2',
            'filters': [
                {'type': 'value',
                 'key': 'State.Name',
                 'value': ['running', 'stopped'],
                 'op': 'in'},
                {'type': 'value',
                 'key': 'InstanceType',
                 'value': ['m4.2xlarge', 'm4.4xlarge'],
                 'op': 'in'},
                ],
            'actions': [
                {'type': 'resize',
                 'restart': True,
                 'default': 'm4.large',
                 'type-map': {
                     'm4.4xlarge': 'm4.2xlarge'}}]
            }, session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources), 3)

        stopped, running = [], []
        for i in resources:
            if i['State']['Name'] == 'running':
                running.append(i['InstanceId'])
            if i['State']['Name'] == 'stopped':
                stopped.append(i['InstanceId'])

        instances = utils.query_instances(
            session_factory(),
            InstanceIds=[r['InstanceId'] for r in resources])

        cur_stopped, cur_running = [], []
        for i in instances:
            if i['State']['Name'] == 'running':
                cur_running.append(i['InstanceId'])
            if i['State']['Name'] == 'stopped':
                cur_stopped.append(i['InstanceId'])

        cur_running.sort()
        running.sort()

        self.assertEqual(cur_stopped, stopped)
        self.assertEqual(cur_running, running)
        instance_types = [i['InstanceType'] for i in instances]
        instance_types.sort()
        self.assertEqual(
            instance_types,
            list(sorted(['m4.large', 'm4.2xlarge', 'm4.2xlarge'])))


class TestStateTransitionAgeFilter(BaseTest):

    def test_ec2_state_transition_age(self):
        session_factory = self.replay_flight_data(
            'test_ec2_state_transition_age_filter'
        )
        policy = self.load_policy({
            'name': 'ec2-state-transition-age',
            'resource': 'ec2',
            'filters': [
                {'State.Name': 'running'},
                {'type': 'state-age',
                 'days': 30}]},
            session_factory=session_factory)
        resources = policy.run()
        #compare stateTransition reason to expected
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['StateTransitionReason'], 'User initiated (2015-11-25 10:11:55 GMT)')
        
    def test_date_parsing(self):
        instance = ec2.StateTransitionAge(None)
        
        # Missing key
        self.assertIsNone(instance.get_resource_date({}))

        # Bad date format
        self.assertRaises(
            ValueError,
            instance.get_resource_date,
            {'StateTransitionReason': "User initiated (201-02-06 17:77:00 GMT)"}
        )
        
        # Won't match regex
        self.assertIsNone(
            instance.get_resource_date({
                'StateTransitionReason': "Server.InternalError"
        }))

        # Test for success
        self.assertEqual(
            instance.get_resource_date({
                'StateTransitionReason': "User initiated (2017-02-06 17:57:00 GMT)"
            }),
            datetime(2017, 2, 6, 17, 57, tzinfo=tz.tzutc())
        )


class TestImageAgeFilter(BaseTest):

    def test_ec2_image_age(self):
        session_factory = self.replay_flight_data(
            'test_ec2_image_age_filter')
        policy = self.load_policy({
            'name': 'ec2-image-age',
            'resource': 'ec2',
            'filters': [
                {'State.Name': 'running'},
                {'type': 'image-age',
                 'days': 30}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestImageFilter(BaseTest):

    def test_ec2_image(self):
        session_factory = self.replay_flight_data(
            'test_ec2_image_filter')
        policy = self.load_policy({
            'name': 'ec2-image',
            'resource': 'ec2',
            'filters': [
                {'type': 'image', 'key': 'Public', 'value': True}
                ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['InstanceId'], 'i-039628786cabe8c16')


class TestInstanceAge(BaseTest):

    # placebo doesn't record tz information
    def test_ec2_instance_age(self):
        session_factory = self.replay_flight_data(
            'test_ec2_instance_age_filter')
        policy = self.load_policy({
            'name': 'ec2-instance-age',
            'resource': 'ec2',
            'filters': [
                {'State.Name': 'running'},
                {'type': 'instance-age',
                 'days': 0}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestTag(BaseTest):

    def test_ec2_tag(self):
        session_factory = self.replay_flight_data(
            'test_ec2_mark')
        policy = self.load_policy({
            'name': 'ec2-test-mark',
            'resource': 'ec2',
            'filters': [
                {'State.Name': 'running'}],
            'actions': [
                {'type': 'tag',
                 'key': 'Testing',
                 'value': 'Testing123'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    def test_ec2_tag_errors(self):
        # Specifying both 'key' and 'tag' is an error
        policy = {
            'name': 'ec2-tag-error',
            'resource': 'ec2',
            'actions': [{
                'type': 'tag',
                'key': 'Testing',
                'tag': 'foo',
                'value': 'TestingError'
            }]
        }
        self.assertRaises(FilterValidationError, self.load_policy, policy)

        # Invalid op for 'mark-for-op' action
        policy = {
            'name': 'ec2-tag-error',
            'resource': 'ec2',
            'actions': [{
                'type': 'mark-for-op',
                'op': 'fake',
            }]
        }
        self.assertRaises(FilterValidationError, self.load_policy, policy)
        
    def test_ec2_untag(self):
        session_factory = self.replay_flight_data(
            'test_ec2_untag')
        policy = self.load_policy({
            'name': 'ec2-test-unmark',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing': 'not-null'}],
            'actions': [
                {'type': 'remove-tag',
                 'tags': ['Testing']}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    def test_ec2_normalize_tag(self):
        session_factory = self.replay_flight_data(
            'test_ec2_normalize_tag')

        policy = self.load_policy({
            'name': 'ec2-test-normalize-tag-lower',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing-lower': 'not-null'}],
            'actions': [
                {'type': 'normalize-tag',
                 'key': 'Testing-lower',
                 'action': 'lower'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

        policy = self.load_policy({
            'name': 'ec2-test-normalize-tag-upper',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing-upper': 'not-null'}],
            'actions': [
                {'type': 'normalize-tag',
                 'key': 'Testing-upper',
                 'action': 'upper'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

        policy = self.load_policy({
            'name': 'ec2-test-normalize-tag-title',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing-title': 'not-null'}],
            'actions': [
                {'type': 'normalize-tag',
                 'key': 'Testing-title',
                 'action': 'title'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

        policy = self.load_policy({
            'name': 'ec2-test-normalize-tag-strip',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing-strip': 'not-null'}],
            'actions': [
                {'type': 'normalize-tag',
                 'key': 'Testing-strip',
                 'action': 'strip',
                 'value': 'blah'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    def test_ec2_rename_tag(self):
        session_factory = self.replay_flight_data(
            'test_ec2_rename_tag')

        policy = self.load_policy({
            'name': 'ec2-rename-start',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing': 'present'}
                ]}, session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 3)

        policy = self.load_policy({
            'name': 'ec2-rename-tag',
            'resource': 'ec2',
            'actions': [{
                'type': 'rename-tag',
                'old_key': 'Testing',
                'new_key': 'Testing1'}]}, session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 3)

        policy = self.load_policy({
            'name': 'ec2-rename-end',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing1': 'present'}
                ]}, session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 3)


class TestStop(BaseTest):

    def test_ec2_stop(self):
        session_factory = self.replay_flight_data(
            'test_ec2_stop')
        policy = self.load_policy({
            'name': 'ec2-test-stop',
            'resource': 'ec2',
            'filters': [
                {'tag:Testing': 'not-null'}],
            'actions': [
                {'type': 'stop'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestStart(BaseTest):

    def test_ec2_start(self):
        session_factory = self.replay_flight_data(
            'test_ec2_start')
        policy = self.load_policy({
            'name': 'ec2-test-start',
            'resource': 'ec2',
            'filters': [],
            'actions': [
                {'type': 'start'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 2)


class TestOr(BaseTest):

    def test_ec2_or_condition(self):
        session_factory = self.replay_flight_data(
            'test_ec2_stop')
        policy = self.load_policy({
            'name': 'ec2-test-snapshot',
            'resource': 'ec2',
            'filters': [
                {"or": [
                    {"tag:Name": "CompileLambda"},
                    {"tag:Name": "Spinnaker"}]}]
        }, session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 2)
        self.assertEqual(
            sorted([r['InstanceId'] for r in resources]),
            [u'i-13413bd7', u'i-1aebf7c0'])


class TestSnapshot(BaseTest):

    def test_ec2_snapshot_no_copy_tags(self):
        session_factory = self.replay_flight_data(
            'test_ec2_snapshot')
        policy = self.load_policy({
            'name': 'ec2-test-snapshot',
            'resource': 'ec2',
            'filters': [
                {'tag:Name': 'CompileLambda'}],
            'actions': [
                {'type': 'snapshot'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    def test_ec2_snapshot_copy_tags(self):
        session_factory = self.replay_flight_data(
            'test_ec2_snapshot')
        policy = self.load_policy({
            'name': 'ec2-test-snapshot',
            'resource': 'ec2',
            'filters': [
                {'tag:Name': 'CompileLambda'}],
            'actions': [
                {'type': 'snapshot', 'copy-tags': ['ASV' 'Testing123']}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class TestEC2QueryFilter(unittest.TestCase):

    def test_parse(self):
        self.assertEqual(QueryFilter.parse([]), [])
        x = QueryFilter.parse(
            [{'instance-state-name': 'running'}])
        self.assertEqual(
            x[0].query(),
            {'Name': 'instance-state-name', 'Values': ['running']})

        self.assertTrue(
            isinstance(
                QueryFilter.parse(
                    [{'tag:ASV': 'REALTIMEMSG'}])[0],
                QueryFilter))

        self.assertRaises(
            ValueError,
            QueryFilter.parse,
            [{'tag:ASV': None}])


class TestTerminate(BaseTest):

    def test_ec2_terminate(self):
        # Test conditions: single running instance, with delete protection
        session_factory = self.replay_flight_data('test_ec2_terminate')
        p = self.load_policy({
            'name': 'ec2-term',
            'resource': 'ec2',
            'filters': [{'InstanceId': 'i-017cf4e2a33b853fe'}],
            'actions': [
                {'type': 'terminate',
                 'force': True}]},
           session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        instances = utils.query_instances(
            session_factory(), InstanceIds=['i-017cf4e2a33b853fe'])
        self.assertEqual(instances[0]['State']['Name'], 'shutting-down')


class TestDefaultVpc(BaseTest):

    def test_ec2_default_vpc(self):
        session_factory = self.replay_flight_data('test_ec2_default_vpc')
        p = self.load_policy(
            {'name': 'ec2-default-filters',
             'resource': 'ec2',
             'filters': [
                 {'type': 'default-vpc'}]},
            config={'region': 'us-west-2'},
            session_factory=session_factory)

        resources = p.run()

        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['InstanceId'], 'i-0bfe468063b02d018')


class TestActions(unittest.TestCase):

    def test_action_construction(self):

        self.assertIsInstance(
            actions.factory('mark', None),
            tags.Tag)

        self.assertIsInstance(
            actions.factory('stop', None),
            ec2.Stop)

        self.assertIsInstance(
            actions.factory('terminate', None),
            ec2.Terminate)


class TestModifySecurityGroupsActionSchema(BaseTest):
    def test_remove_dependencies(self):
        policy = {
            'name': 'remove-with-no-isolation-or-add',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups', 'remove': 'matched'}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, data=policy, validate=True)

    def test_invalid_remove_params(self):
        # basestring invalid
        policy = {
            'name': 'remove-with-incorrect-param-string',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups', 'remove': 'none'}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, data=policy, validate=True)

        # list - one valid, one invalid
        policy = {
            'name': 'remove-with-incorrect-param-list',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups', 'remove': [
                    'invalid-sg', 'sg-abcd1234']}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, policy, validate=True)

    def test_invalid_add_params(self):
        # basestring invalid
        policy = {
            'name': 'add-with-incorrect-param-string',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups', 'add': 'none'},
                {'type': 'modify-security-groups', 'add': [
                    'invalid-sg', 'sg-abcd1234']}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, data=policy, validate=True)

    def test_invalid_isolation_group_params(self):
        policy = {
            'name': 'isolation-group-with-incorrect-param-string',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups', 'isolation-group': 'none'}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, data=policy, validate=True)

        # list - one valid, one invalid
        policy = {
            'name': 'isolation-group-with-incorrect-param-list',
            'resource': 'ec2',
            'actions': [
                {'type': 'modify-security-groups',
                 'isolation-group': ['invalid-sg', 'sg-abcd1234']}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, data=policy, validate=True)


class TestModifySecurityGroupAction(BaseTest):
    def test_security_group_type(self):
        # Test conditions:
        #   - running two instances; one with TestProductionInstanceProfile
        #     and one with none
        #   - security group named TEST-PROD-ONLY-SG exists in VPC and is
        #     attached to both test instances
        session_factory = self.replay_flight_data(
            'test_ec2_security_group_filter')

        # Catch on anything that uses the *PROD-ONLY* security groups but isn't in a prod role
        policy = self.load_policy({
            'name': 'restrict-sensitive-sg',
            'resource': 'ec2',
            'filters': [
                {'or': [
                    {'and': [
                        {'type': 'value', 'key': 'IamInstanceProfile.Arn',
                         'value': '(?!.*TestProductionInstanceProfile)(.*)',
                         'op': 'regex'},
                        {'type': 'value', 'key': 'IamInstanceProfile.Arn',
                         'value': 'not-null'}
                    ]},
                    {'type': 'value', 'key': 'IamInstanceProfile',
                     'value': 'absent'}
                ]},
                {'type': 'security-group', 'key': 'GroupName',
                 'value': '(.*PROD-ONLY.*)', 'op': 'regex'},

            ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['InstanceId'], 'i-0dd3919bc5bac1ea8')

    def test_security_group_modify_groups_action(self):
        # Test conditions:
        #   - running two instances; one with TestProductionInstanceProfile
        #     and one with none
        #   - security group named TEST-PROD-ONLY-SG exists in VPC and is
        #     attached to both test instances
        session_factory = self.replay_flight_data(
            'test_ec2_modify_groups_action')
        client = session_factory().client('ec2')

        default_sg_id = client.describe_security_groups(
            GroupNames=[
                'default',
            ]
        )['SecurityGroups'][0]['GroupId']


        # Catch on anything that uses the *PROD-ONLY* security groups but isn't in a prod role
        policy = self.load_policy({
            'name': 'remove-sensitive-sg',
            'resource': 'ec2',
            'filters': [
                {'or': [
                    {'and': [
                        {'type': 'value', 'key': 'IamInstanceProfile.Arn',
                         'value': '(?!.*TestProductionInstanceProfile)(.*)',
                         'op': 'regex'},
                        {'type': 'value', 'key': 'IamInstanceProfile.Arn',
                         'value': 'not-null'}
                    ]},
                    {'type': 'value', 'key': 'IamInstanceProfile',
                     'value': 'absent'}
                ]},
                {'type': 'security-group', 'key': 'GroupName',
                 'value': '(.*PROD-ONLY.*)', 'op': 'regex'}],
            'actions': [
                {'type': 'modify-security-groups', 'remove': 'matched',
                 'isolation-group': default_sg_id}]
            },
            session_factory=session_factory)
        before_action_resources = policy.run()
        after_action_resources = policy.run()
        self.assertEqual(len(before_action_resources), 1)
        self.assertEqual(
            before_action_resources[0]['InstanceId'], 'i-0dd3919bc5bac1ea8')
        self.assertEqual(len(after_action_resources), 0)

    def test_invalid_modify_groups_schema(self):
        policy = {
            'name': 'invalid-modify-security-groups-action',
            'resource': 'ec2',
            'filters': [],
            'actions': [
                {'type': 'modify-security-groups', 'change': 'matched'}
            ]
        }
        self.assertRaises(
            ValidationError, self.load_policy, policy, validate=True)

    def test_ec2_add_security_groups(self):
        # Test conditions:
        #   - running one instance with TestProductionInstanceProfile
        #   - security group named TEST-PROD-ONLY-SG exists in VPC and
        #     is attached to test instance
        #   - security group with id sg-8a4b64f7 exists in VPC and is selected
        #     in a policy to be attached
        session_factory = self.replay_flight_data(
            'test_ec2_add_security_groups')
        policy = self.load_policy({
            'name': 'add-sg-to-prod-instances',
            'resource': 'ec2',
            'filters': [
                {'type': 'value', 'key': 'IamInstanceProfile.Arn',
                 'value': '(.*TestProductionInstanceProfile)', 'op': 'regex'}
            ],
            'actions': [
                {'type': 'modify-security-groups', 'add': 'sg-8a4b64f7'}
            ]
        },
        session_factory=session_factory)

        first_resources = policy.run()
        self.assertEqual(len(
            first_resources[0]['NetworkInterfaces'][0]['Groups']), 1)
        second_resources = policy.run()
        self.assertEqual(len(
            second_resources[0]['NetworkInterfaces'][0]['Groups']), 2)


class TestFilter(BaseTest):
    
    def test_not_filter(self):
        # This test is to get coverage for the `not` filter's process_set method
        session_factory = self.replay_flight_data(
            'test_ec2_not_filter')

        policy = self.load_policy({
            'name': 'list-ec2-test-not',
            'resource': 'ec2',
            'filters': [{
                'not': [
                    {'InstanceId': 'i-036ee05e8c2ca83b3'}
                ]
            }]
        },
        session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources), 2)

        policy = self.load_policy({
            'name': 'list-ec2-test-not',
            'resource': 'ec2',
            'filters': [{
                'not': [{
                    'or': [
                        {'InstanceId': 'i-036ee05e8c2ca83b3'},
                        {'InstanceId': 'i-03d8207d8285cbf53'}
                    ]
                }]
            }]
        },
        session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources), 1)



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Application Load Balancers
"""
import logging

from collections import defaultdict
from c7n.actions import ActionRegistry, BaseAction
from c7n.filters import (
    Filter, FilterRegistry, FilterValidationError, DefaultVpcBase, ValueFilter)
import c7n.filters.vpc as net_filters
from c7n import tags
from c7n.manager import resources
from c7n.query import QueryResourceManager
from c7n.utils import local_session, chunks, type_schema, get_retry

log = logging.getLogger('custodian.app-elb')

filters = FilterRegistry('app-elb.filters')
actions = ActionRegistry('app-elb.actions')

filters.register('tag-count', tags.TagCountFilter)
filters.register('marked-for-op', tags.TagActionFilter)


@resources.register('app-elb')
class AppELB(QueryResourceManager):
    """Resource manager for v2 ELBs (AKA ALBs).
    """

    class resource_type(object):

        service = 'elbv2'
        type = 'app-elb'
        enum_spec = ('describe_load_balancers', 'LoadBalancers', None)
        name = 'LoadBalancerName'
        id = 'LoadBalancerArn'
        filter_name = None
        filter_type = None
        dimension = None
        date = 'CreatedTime'
        config_type = 'AWS::ElasticLoadBalancingV2::LoadBalancer'

    filter_registry = filters
    action_registry = actions
    retry = staticmethod(get_retry(('Throttling',)))

    @classmethod
    def get_permissions(cls):
        # override as the service is not the iam prefix
        return ("elasticloadbalancing:DescribeLoadBalancers",
                "elasticloadbalancing:DescribeTags")

    def augment(self, albs):
        _describe_appelb_tags(
            albs, self.session_factory,
            self.executor_factory, self.retry)

        return albs


def _describe_appelb_tags(albs, session_factory, executor_factory, retry):
    def _process_tags(alb_set):
        client = local_session(session_factory).client('elbv2')
        alb_map = {alb['LoadBalancerArn']: alb for alb in alb_set}

        results = retry(client.describe_tags, ResourceArns=alb_map.keys())
        for tag_desc in results['TagDescriptions']:
            if ('ResourceArn' in tag_desc and
                    tag_desc['ResourceArn'] in alb_map):
                alb_map[tag_desc['ResourceArn']]['Tags'] = tag_desc['Tags']

    with executor_factory(max_workers=2) as w:
        list(w.map(_process_tags, chunks(albs, 20)))


def _add_appelb_tags(albs, session_factory, ts):
    client = local_session(session_factory).client('elbv2')
    client.add_tags(
        ResourceArns=[alb['LoadBalancerArn'] for alb in albs],
        Tags=ts)


def _remove_appelb_tags(albs, session_factory, tag_keys):
    client = local_session(session_factory).client('elbv2')
    client.remove_tags(
        ResourceArns=[alb['LoadBalancerArn'] for alb in albs],
        TagKeys=tag_keys)


@filters.register('security-group')
class SecurityGroupFilter(net_filters.SecurityGroupFilter):

    RelatedIdsExpression = "SecurityGroups[]"


@filters.register('subnet')
class SubnetFilter(net_filters.SubnetFilter):

    RelatedIdsExpression = "AvailabilityZones[].SubnetId"


@actions.register('mark-for-op')
class AppELBMarkForOpAction(tags.TagDelayedAction):
    """Action to create a delayed action on an ELB to start at a later date

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-failed-mark-for-op
                resource: app-elb
                filters:
                  - "tag:custodian_elb_cleanup": absent
                  - State: failed
                actions:
                  - type: mark-for-op
                    tag: custodian_elb_cleanup
                    msg: "AppElb failed: {op}@{action_date}"
                    op: delete
                    days: 1
    """

    batch_size = 1
    permissions = ("elasticloadbalancing:AddTags",)

    def process_resource_set(self, resource_set, ts):
        _add_appelb_tags(
            resource_set,
            self.manager.session_factory,
            ts)


@actions.register('tag')
class AppELBTagAction(tags.Tag):
    """Action to create tag/tags on an ELB

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-create-required-tag
                resource: app-elb
                filters:
                  - "tag:RequiredTag": absent
                actions:
                  - type: tag
                    key: RequiredTag
                    value: RequiredValue
    """

    batch_size = 1
    permissions = ("elasticloadbalancing:AddTags",)

    def process_resource_set(self, resource_set, ts):
        _add_appelb_tags(
            resource_set,
            self.manager.session_factory,
            ts)


@actions.register('remove-tag')
class AppELBRemoveTagAction(tags.RemoveTag):
    """Action to remove tag/tags from an ELB

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-delete-expired-tag
                resource: app-elb
                filters:
                  - "tag:ExpiredTag": present
                actions:
                  - type: remove-tag
                    tags: ["ExpiredTag"]
    """

    batch_size = 1
    permissions = ("elasticloadbalancing:RemoveTags",)

    def process_resource_set(self, resource_set, tag_keys):
        _remove_appelb_tags(
            resource_set,
            self.manager.session_factory,
            tag_keys)


@actions.register('delete')
class AppELBDeleteAction(BaseAction):
    """Action to delete an ELB

    To avoid unwanted deletions of ELB, it is recommended to apply a filter
    to the rule

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-delete-failed-elb
                resource: app-elb
                filters:
                  - State: failed
                actions:
                  - delete
    """

    schema = type_schema('delete', force={'type': 'boolean'})
    permissions = (
        "elasticloadbalancing:DeleteLoadBalancer",
        "elasticloadbalancing:ModifyLoadBalancerAttributes",)

    def process(self, load_balancers):
        with self.executor_factory(max_workers=2) as w:
            list(w.map(self.process_alb, load_balancers))

    def process_alb(self, alb):
        client = local_session(self.manager.session_factory).client('elbv2')
        try:
            if self.data.get('force'):
                client.modify_load_balancer_attributes(
                    LoadBalancerArn=alb['LoadBalancerArn'],
                    Attributes=[{
                        'Key': 'deletion_protection.enabled',
                        'Value': 'false',
                    }])
            self.manager.retry(
                client.delete_load_balancer, LoadBalancerArn=alb['LoadBalancerArn'])
        except Exception as e:
            if e.response['Error']['Code'] in ['OperationNotPermitted',
                                               'LoadBalancerNotFound']:
                self.log.warning(
                    "Exception trying to delete ALB: %s error: %s",
                    alb['LoadBalancerArn'], e)
                return
            raise


class AppELBListenerFilterBase(object):
    """ Mixin base class for filters that query LB listeners.
    """
    permissions = ("elasticloadbalancing:DescribeListeners",)

    def initialize(self, albs):
        def _process_listeners(alb):
            client = local_session(
                self.manager.session_factory).client('elbv2')
            results = client.describe_listeners(
                LoadBalancerArn=alb['LoadBalancerArn'])
            self.listener_map[alb['LoadBalancerArn']] = results['Listeners']

        self.listener_map = defaultdict(list)
        with self.manager.executor_factory(max_workers=2) as w:
            list(w.map(_process_listeners, albs))


class AppELBAttributeFilterBase(object):
    """ Mixin base class for filters that query LB attributes.
    """

    def initialize(self, albs):
        def _process_attributes(alb):
            if 'Attributes' not in alb:
                client = local_session(
                    self.manager.session_factory).client('elbv2')
                results = client.describe_load_balancer_attributes(
                    LoadBalancerArn=alb['LoadBalancerArn'])
                alb['Attributes'] = results['Attributes']

        with self.manager.executor_factory(max_workers=2) as w:
            list(w.map(_process_attributes, albs))


class AppELBTargetGroupFilterBase(object):
    """ Mixin base class for filters that query LB target groups.
    """

    def initialize(self, albs):
        self.target_group_map = defaultdict(list)
        target_groups = self.manager.get_resource_manager(
            'app-elb-target-group').resources()
        for target_group in target_groups:
            for load_balancer_arn in target_group['LoadBalancerArns']:
                self.target_group_map[load_balancer_arn].append(target_group)


@filters.register('listener')
class AppELBListenerFilter(ValueFilter, AppELBListenerFilterBase):
    """Filter ALB based on matching listener attributes"""

    schema = type_schema('listener', rinherit=ValueFilter.schema)
    permissions = ("elasticloadbalancing:DescribeLoadBalancerAttributes",)

    def process(self, albs, event=None):
        self.initialize(albs)
        return super(AppELBListenerFilter, self).process(albs, event)

    def __call__(self, alb):
        matched = []
        for listener in self.listener_map[alb['LoadBalancerArn']]:
            if self.match(listener):
                matched.append(listener)
        if not matched:
            return False
        alb['c7n:MatchedListeners'] = matched
        return True


@actions.register('modify-listener')
class AppELBModifyListenerPolicy(BaseAction):
    """Action to modify the policy for an App ELB

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-modify-listener
                resource: app-elb
                filters:
                  - type: listener
                    key: Protocol
                    value: HTTP
                actions:
                  - type: modify-listener
                    protocol: HTTPS
                    sslpolicy: "ELBSecurityPolicy-TLS-1-2-2017-01"
                    certificate: "arn:aws:acm:region:123456789012:certificate/12345678-\
                    1234-1234-1234-123456789012"
    """

    schema = type_schema(
        'modify-listener',
        port={'type': 'integer'},
        protocol={'enum': ['HTTP', 'HTTPS']},
        sslpolicy={'type': 'string'},
        certificate={'type': 'string'}
    )

    permissions = ("elasticloadbalancing:ModifyListener",)

    def validate(self):
        for f in self.manager.data.get('filters', ()):
            if 'listener' in f.get('type', ()):
                return self
        raise FilterValidationError(
            "modify-listener action requires the listener filter")

    def process(self, load_balancers):
        args = {}
        if 'port' in self.data:
            args['Port'] = self.data.get('port')
        if 'protocol' in self.data:
            args['Protocol'] = self.data.get('protocol')
        if 'sslpolicy' in self.data:
            args['SslPolicy'] = self.data.get('sslpolicy')
        if 'certificate' in self.data:
            args['Certificates'] = [{'CertificateArn': self.data.get('certificate')}]
        with self.executor_factory(max_workers=2) as w:
            list(w.map(self.process_alb, load_balancers, [args]))

    def process_alb(self, alb, args):
        client = local_session(self.manager.session_factory).client('elbv2')
        for matched_listener in alb.get('c7n:MatchedListeners', ()):
            client.modify_listener(
                ListenerArn=matched_listener['ListenerArn'],
                **args)


@filters.register('healthcheck-protocol-mismatch')
class AppELBHealthCheckProtocolMismatchFilter(Filter,
                                              AppELBTargetGroupFilterBase):
    """Filter AppELBs with mismatched health check protocols

    A mismatched health check protocol is where the protocol on the target group
    does not match the load balancer health check protocol

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-healthcheck-mismatch
                resource: app-elb
                filters:
                  - healthcheck-protocol-mismatch
    """

    schema = type_schema('healthcheck-protocol-mismatch')
    permissions = ("elasticloadbalancing:DescribeTargetGroups",)

    def process(self, albs, event=None):
        def _healthcheck_protocol_mismatch(alb):
            for target_group in self.target_group_map[alb['LoadBalancerArn']]:
                if (target_group['Protocol'] !=
                        target_group['HealthCheckProtocol']):
                    return True

            return False

        self.initialize(albs)
        return [alb for alb in albs if _healthcheck_protocol_mismatch(alb)]


@filters.register('target-group')
class AppELBTargetGroupFilter(ValueFilter, AppELBTargetGroupFilterBase):
    """Filter ALB based on matching target group value"""

    schema = type_schema('target-group', rinherit=ValueFilter.schema)
    permissions = ("elasticloadbalancing:DescribeTargetGroups",)

    def process(self, albs, event=None):
        self.initialize(albs)
        return super(AppELBTargetGroupFilter, self).process(albs, event)

    def __call__(self, alb):
        target_groups = self.target_group_map[alb['LoadBalancerArn']]
        return self.match(target_groups)


@filters.register('default-vpc')
class AppELBDefaultVpcFilter(DefaultVpcBase):
    """Filter all ELB that exist within the default vpc

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-in-default-vpc
                resource: app-elb
                filters:
                  - default-vpc
    """

    schema = type_schema('default-vpc')

    def __call__(self, alb):
        return alb.get('VpcId') and self.match(alb.get('VpcId')) or False


@resources.register('app-elb-target-group')
class AppELBTargetGroup(QueryResourceManager):
    """Resource manager for v2 ELB target groups.
    """

    class resource_type(object):

        service = 'elbv2'
        type = 'app-elb-target-group'
        enum_spec = ('describe_target_groups', 'TargetGroups', None)
        name = 'TargetGroupName'
        id = 'TargetGroupArn'
        filter_name = None
        filter_type = None
        dimension = None
        date = None

    filter_registry = FilterRegistry('app-elb-target-group.filters')
    action_registry = ActionRegistry('app-elb-target-group.actions')
    retry = staticmethod(get_retry(('Throttling',)))

    filter_registry.register('tag-count', tags.TagCountFilter)
    filter_registry.register('marked-for-op', tags.TagActionFilter)

    @classmethod
    def get_permissions(cls):
        # override as the service is not the iam prefix
        return ("elasticloadbalancing:DescribeTargetGroups",
                "elasticloadbalancing:DescribeTags")

    def augment(self, target_groups):
        def _describe_target_group_health(target_group):
            client = local_session(self.session_factory).client('elbv2')
            result = client.describe_target_health(
                TargetGroupArn=target_group['TargetGroupArn'])
            target_group['TargetHealthDescriptions'] = result[
                'TargetHealthDescriptions']

        with self.executor_factory(max_workers=2) as w:
            list(w.map(_describe_target_group_health, target_groups))

        _describe_target_group_tags(
            target_groups, self.session_factory,
            self.executor_factory, self.retry)
        return target_groups


def _describe_target_group_tags(target_groups, session_factory,
                                executor_factory, retry):
    def _process_tags(target_group_set):
        client = local_session(session_factory).client('elbv2')
        target_group_map = {
            target_group['TargetGroupArn']:
                target_group for target_group in target_group_set
        }

        results = retry(
            client.describe_tags,
            ResourceArns=target_group_map.keys())
        for tag_desc in results['TagDescriptions']:
            if ('ResourceArn' in tag_desc and
                    tag_desc['ResourceArn'] in target_group_map):
                target_group_map[
                    tag_desc['ResourceArn']
                ]['Tags'] = tag_desc['Tags']

    with executor_factory(max_workers=2) as w:
        list(w.map(_process_tags, chunks(target_groups, 20)))


def _add_target_group_tags(target_groups, session_factory, ts):
    client = local_session(session_factory).client('elbv2')
    client.add_tags(
        ResourceArns=[
            target_group['TargetGroupArn'] for target_group in target_groups
        ],
        Tags=ts)


def _remove_target_group_tags(target_groups, session_factory, tag_keys):
    client = local_session(session_factory).client('elbv2')
    client.remove_tags(
        ResourceArns=[
            target_group['TargetGroupArn'] for target_group in target_groups
        ],
        TagKeys=tag_keys)


@AppELBTargetGroup.action_registry.register('mark-for-op')
class AppELBTargetGroupMarkForOpAction(tags.TagDelayedAction):
    """Action to specify a delayed action on an ELB target group"""

    batch_size = 1
    permissions = ("elasticloadbalancing:AddTags",)

    def process_resource_set(self, resource_set, ts):
        _add_target_group_tags(
            resource_set,
            self.manager.session_factory,
            ts)


@AppELBTargetGroup.action_registry.register('tag')
class AppELBTargetGroupTagAction(tags.Tag):
    """Action to create tag/tags on an ELB target group

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-targetgroup-add-required-tag
                resource: app-elb-target-group
                filters:
                  - "tag:RequiredTag": absent
                actions:
                  - type: tag
                    key: RequiredTag
                    value: RequiredValue
    """

    batch_size = 1
    permissions = ("elasticloadbalancing:AddTags",)

    def process_resource_set(self, resource_set, ts):
        _add_target_group_tags(
            resource_set,
            self.manager.session_factory,
            ts)


@AppELBTargetGroup.action_registry.register('remove-tag')
class AppELBTargetGroupRemoveTagAction(tags.RemoveTag):
    """Action to remove tag/tags from ELB target group

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-targetgroup-remove-expired-tag
                resource: app-elb-target-group
                filters:
                  - "tag:ExpiredTag": present
                actions:
                  - type: remove-tag
                    tags: ["ExpiredTag"]
    """

    batch_size = 1
    permissions = ("elasticloadbalancing:RemoveTags",)

    def process_resource_set(self, resource_set, tag_keys):
        _remove_target_group_tags(
            resource_set,
            self.manager.session_factory,
            tag_keys)


@AppELBTargetGroup.filter_registry.register('default-vpc')
class AppELBTargetGroupDefaultVpcFilter(DefaultVpcBase):
    """Filter all application elb target groups within the default vpc

    :example:

        .. code-block: yaml

            policies:
              - name: appelb-targetgroups-default-vpc
                resource: app-elb-target-group
                filters:
                  - default-vpc
    """

    schema = type_schema('default-vpc')

    def __call__(self, target_group):
        return (target_group.get('VpcId') and
                self.match(target_group.get('VpcId')) or False)



# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from common import BaseTest


class CodeCommit(BaseTest):

    def test_query_repos(self):
        factory = self.replay_flight_data('test_codecommit')
        p = self.load_policy({
            'name': 'get-repos',
            'resource': 'codecommit',
        }, session_factory=factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(
            resources[0]['cloneUrlSsh'],
            "ssh://git-codecommit.us-east-2.amazonaws.com/v1/repos/custodian-config-repo")


class CodeBuild(BaseTest):

    def test_query_builds(self):
        factory = self.replay_flight_data('test_codebuild')
        p = self.load_policy({
            'name': 'get-builders',
            'resource': 'codebuild',
        }, session_factory=factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

        self.assertEqual(
            resources[0]['environment'],
            {u'computeType': u'BUILD_GENERAL1_SMALL',
             u'environmentVariables': [],
             u'image': u'aws/codebuild/python:2.7.12',
             u'type': u'LINUX_CONTAINER'})


class CodePipeline(BaseTest):

    def test_query_pipeline(self):
        factory = self.replay_flight_data('test_codepipeline')
        p = self.load_policy({
            'name': 'get-pipes',
            'resource': 'codepipeline',
        }, session_factory=factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(len(resources[0]['stages']), 2)



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from common import BaseTest


class SimpleDB(BaseTest):

    def test_delete(self):
        session_factory = self.replay_flight_data('test_simpledb_delete')
        p = self.load_policy({
            'name': 'sdb-del',
            'resource': 'simpledb',
            'filters': [
                {'DomainName': 'supersuper'}],
            'actions': ['delete']
        }, session_factory=session_factory)

        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['DomainName'], 'supersuper')
        extant_domains = session_factory().client(
            'sdb').list_domains()['DomainNames']
        self.assertTrue(
            resources[0]['DomainName'] not in extant_domains)

    def test_simpledb(self):
        session_factory = self.replay_flight_data('test_simpledb_query')
        p = self.load_policy({
            'name': 'sdbtest',
            'resource': 'simpledb'},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['DomainName'], 'devtest')



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
import os

from dateutil import zoneinfo

from mock import mock

from common import BaseTest, instance

from c7n.filters import FilterValidationError
from c7n.filters.offhours import OffHour, OnHour, ScheduleParser, Time


# Per http://blog.xelnor.net/python-mocking-datetime/
# naive implementation has issues with pypy

real_datetime_class = datetime.datetime


def mock_datetime_now(tgt, dt):

    class DatetimeSubclassMeta(type):
        @classmethod
        def __instancecheck__(mcs, obj):
            return isinstance(obj, real_datetime_class)

    class BaseMockedDatetime(real_datetime_class):
        target = tgt

        @classmethod
        def now(cls, tz=None):
            return cls.target.replace(tzinfo=tz)

        @classmethod
        def utcnow(cls):
            return cls.target

        # Python2 & Python3 compatible metaclass

    MockedDatetime = DatetimeSubclassMeta(
        'datetime', (BaseMockedDatetime,), {})
    return mock.patch.object(dt, 'datetime', MockedDatetime)


class OffHoursFilterTest(BaseTest):
    """[off|on] hours testing"""

    def test_offhours_records(self):
        session_factory = self.replay_flight_data('test_offhours_records')
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=8, day=14, hour=19, minute=00)

        with mock_datetime_now(t, datetime):
            p = self.load_policy({
                'name': 'offhours-records',
                'resource': 'ec2',
                'filters': [
                    {'State.Name': 'running'},
                    {'type': 'offhour',
                     'offhour': 19,
                     'tag': 'custodian_downtime',
                     'default_tz': 'est',
                     'weekends': False}]
            }, session_factory=session_factory)
            resources = p.run()
        self.assertEqual(resources, [])
        with open(os.path.join(
                p.options['output_dir'],
                'offhours-records',
                'parse_errors.json')) as fh:
            data = json.load(fh)
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0][0], 'i-0ee3a9bc2eeed269f')
            self.assertEqual(data[0][1], 'off=[m-f,8];on=[n-f,5];pz=est')
        with open(os.path.join(
                p.options['output_dir'],
                'offhours-records',
                'opted_out.json')) as fh:
            data = json.load(fh)
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0]['InstanceId'], 'i-0a619b58a7e704a9f')

    def test_validate(self):
        self.assertRaises(
            FilterValidationError, OffHour({'default_tz': 'zmta'}).validate)
        self.assertRaises(
            FilterValidationError, OffHour({'offhour': 25}).validate)
        i = OffHour({})
        self.assertEqual(i.validate(), i)

    def test_process(self):
        f = OffHour({'opt-out': True})
        instances = [
            instance(Tags=[]),
            instance(
                Tags=[{'Key': 'maid_offhours', 'Value': 'off'}]),
            instance(
                Tags=[
                    {'Key': 'maid_offhours',
                     'Value': "off=(m-f,5);zebrablue,on=(t-w,5)"}])]
        t = datetime.datetime(
            year=2015, month=12, day=1, hour=19, minute=5,
            tzinfo=zoneinfo.gettz('America/New_York'))
        with mock_datetime_now(t, datetime):
            self.assertEqual(f.process(instances), [instances[0]])

    def test_opt_out_behavior(self):
        # Some users want to match based on policy filters to
        # a resource subset with default opt out behavior
        t = datetime.datetime(
            year=2015, month=12, day=1, hour=19, minute=5,
            tzinfo=zoneinfo.gettz('America/New_York'))
        f = OffHour({'opt-out': True})

        with mock_datetime_now(t, datetime):
            i = instance(Tags=[])
            self.assertEqual(f(i), True)
            i = instance(
                Tags=[{'Key': 'maid_offhours', 'Value': 'off'}])
            self.assertEqual(f(i), False)
            self.assertEqual(f.opted_out, [i])

    def test_opt_in_behavior(self):
        # Given the addition of opt out behavior, verify if its
        # not configured that we don't touch an instance that
        # has no downtime tag
        t = datetime.datetime(
            year=2015, month=12, day=1, hour=19, minute=5,
            tzinfo=zoneinfo.gettz('America/New_York'))
        i = instance(Tags=[])
        f = OffHour({})

        with mock_datetime_now(t, datetime):
            self.assertEqual(f(i), False)
            t = datetime.datetime(
                year=2015, month=12, day=1, hour=7, minute=5,
                tzinfo=zoneinfo.gettz('America/New_York'))
            f = OnHour({})
            self.assertEqual(f(i), False)

    def xtest_time_match_stops_after_skew(self):
        hour = 7
        t = datetime.datetime(
            year=2015, month=12, day=1, hour=hour, minute=5,
            tzinfo=zoneinfo.gettz('America/New_York'))
        i = instance(Tags=[
            {'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OnHour({'skew': 1})
        results = []

        with mock_datetime_now(t, datetime) as dt:
            for n in range(0, 4):
                dt.target = t.replace(hour=hour + n)
                results.append(f(i))
        self.assertEqual(results, [True, True, False, False])

    def test_resource_schedule_error(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2015, month=12, day=1, hour=19, minute=5)
        f = OffHour({})
        f.process_resource_schedule = lambda: False
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[
                {'Key': 'maid_offhours', 'Value': 'tz=est'}])
            self.assertEqual(f(i), False)

    def test_time_filter_usage_errors(self):
        self.assertRaises(NotImplementedError, Time, {})

    def test_everyday_onhour(self):
        # weekends on means we match times on the weekend
        start_day = 14  # sunday
        t = datetime.datetime(
            year=2016, day=start_day, month=8, hour=7, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OnHour({'weekends': False})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(7):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(results, [True] * 7)

    def test_everyday_offhour(self):
        # weekends on means we match times on the weekend
        start_day = 14  # sunday
        t = datetime.datetime(
            year=2016, day=start_day, month=8, hour=19, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OffHour({'weekends': False})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(7):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(results, [True] * 7)

    def test_weekends_only_onhour_support(self):
        # start day is a sunday, weekend only means we only start
        # on monday morning.
        start_day = 14
        t = datetime.datetime(
            year=2016, day=start_day, month=8, hour=7, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OnHour({'weekends-only': True})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(7):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(results, [
            False, True, False, False, False, False, False])

    def test_weekends_only_offhour_support(self):
        # start day is a sunday, weekend only means we only stop
        # on friday evening.
        start_day = 14
        t = datetime.datetime(
            year=2016, day=start_day, month=8, hour=7, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OnHour({'weekends-only': True})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(7):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(results, [
            False, True, False, False, False, False, False])

    def test_onhour_weekend_support(self):
        start_day = 14
        t = datetime.datetime(
            year=2016, day=start_day, month=2, hour=19, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OffHour({'weekends-only': True})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(7):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(
            results,
            [False, False, False, False, False, True, False])

    def test_offhour_weekend_support(self):
        start_day = 26
        t = datetime.datetime(
            year=2016, day=start_day, month=2, hour=19, minute=20)
        i = instance(Tags=[{'Key': 'maid_offhours', 'Value': 'tz=est'}])
        f = OffHour({})
        results = []
        with mock_datetime_now(t, datetime) as dt:
            for n in range(0, 4):
                dt.target = t.replace(day=start_day + n)
                results.append(f(i))
        self.assertEqual(results, [True, False, False, True])

    def test_current_time_test(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2015, month=12, day=1, hour=19, minute=5)
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[
                {'Key': 'maid_offhours', 'Value': 'tz=est'}])
            f = OffHour({})
            p = f.get_tag_value(i)
            self.assertEqual(p, 'tz=est')
            tz = f.get_tz('est')
            self.assertTrue(str(tz) in (
                "tzfile('US/Eastern')",
                "tzfile('America/New_York')"))
            self.assertEqual(
                datetime.datetime.now(tz), t)
            self.assertEqual(t.hour, 19)

    def test_offhours_real_world_values(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2015, month=12, day=1, hour=19, minute=5)
        with mock_datetime_now(t, datetime):
            results = [OffHour({})(i) for i in [
                instance(Tags=[
                    {'Key': 'maid_offhours', 'Value': ''}]),
                instance(Tags=[
                    {'Key': 'maid_offhours', 'Value': '"Offhours tz=ET"'}]),
                instance(Tags=[
                    {'Key': 'maid_offhours', 'Value': 'Offhours tz=PT'}])]]
            # unclear what this is really checking
            self.assertEqual(results, [True, True, True])

    def test_offhours_get_value(self):
        off = OffHour({'default_tz': 'ct'})
        i = instance(Tags=[
            {'Key': 'maid_offhours', 'Value': 'Offhours tz=PT'}])
        self.assertEqual(off.get_tag_value(i), "offhours tz=pt")
        self.assertFalse(off.parser.has_resource_schedule(
            off.get_tag_value(i), 'off'))
        self.assertTrue(off.parser.keys_are_valid(
            off.get_tag_value(i)))
        self.assertEqual(off.parser.raw_data(
            off.get_tag_value(i)), {'tz': 'pt'})

    def test_offhours(self):
        t = datetime.datetime(year=2015, month=12, day=1, hour=19, minute=5,
                              tzinfo=zoneinfo.gettz('America/New_York'))
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[
                {'Key': 'maid_offhours', 'Value': 'tz=est'}])
            self.assertEqual(OffHour({})(i), True)

    def test_onhour(self):
        t = datetime.datetime(year=2015, month=12, day=1, hour=7, minute=5,
                              tzinfo=zoneinfo.gettz('America/New_York'))
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[
                {'Key': 'maid_offhours', 'Value': 'tz=est'}])
            self.assertEqual(OnHour({})(i), True)
            self.assertEqual(OnHour({'onhour': 8})(i), False)

    def test_cant_parse_tz(self):
        i = instance(Tags=[
            {'Key': 'maid_offhours', 'Value': 'tz=evt'}])
        self.assertEqual(OffHour({})(i), False)

    def test_custom_offhours(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=5, day=26, hour=19, minute=00)
        results = []

        with mock_datetime_now(t, datetime):
            for i in [instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,19);on=(m-f,7);tz=et'}]),
                      instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,20);on=(m-f,7);tz=et'}])]:
                results.append(OffHour({})(i))
            self.assertEqual(results, [True, False])

    def test_custom_onhours(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=5, day=26, hour=7, minute=00)
        results = []

        with mock_datetime_now(t, datetime):
            for i in [instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,19);on=(m-f,7);tz=et'}]),
                      instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,20);on=(m-f,9);tz=et'}])]:
                results.append(OnHour({})(i))
            self.assertEqual(results, [True, False])

    def test_custom_bad_tz(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=5, day=26, hour=7, minute=00)
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,19);on=(m-f,7);tz=et'}])
            self.assertEqual(OnHour({})(i), True)

            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,20);on=(m-f,7);tz=abc'}])
            self.assertEqual(OnHour({})(i), False)

    def test_custom_bad_hours(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=5, day=26, hour=19, minute=00)
        # default error handling is to exclude the resource

        with mock_datetime_now(t, datetime):
            # This isn't considered a bad value, its basically omitted.
            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=();tz=et'}])
            self.assertEqual(OffHour({})(i), False)

            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,90);on=(m-f,7);tz=et'}])
            # malformed value
            self.assertEqual(OffHour({})(i), False)

        t = t.replace(year=2016, month=5, day=26, hour=13, minute=00)
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=();tz=et'}])
            # will go to default values, but not work due to default time
            self.assertEqual(OffHour({})(i), False)

            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'off=(m-f,90);on=(m-f,7);tz=et'}])
            self.assertEqual(OffHour({})(i), False)

    def test_tz_only(self):
        t = datetime.datetime.now(zoneinfo.gettz('America/New_York'))
        t = t.replace(year=2016, month=5, day=26, hour=7, minute=00)
        with mock_datetime_now(t, datetime):
            i = instance(Tags=[{'Key': 'maid_offhours',
                                'Value': 'tz=est'}])
            self.assertEqual(OnHour({})(i), True)


class ScheduleParserTest(BaseTest):
    # table style test
    # list of (tag value, parse result)
    table = [

        ################
        # Standard cases
        ('off=(m-f,10);on=(m-f,7);tz=et',
         {'off': [{'days': [0, 1, 2, 3, 4], 'hour': 10}],
          'on': [{'days': [0, 1, 2, 3, 4], 'hour': 7}],
          'tz': 'et'}),
        ("off=[(m-f,9)];on=(m-s,10);tz=pt",
         {'off': [{'days': [0, 1, 2, 3, 4], 'hour': 9}],
          'on': [{'days': [0, 1, 2, 3, 4, 5], 'hour': 10}],
          'tz': 'pt'}),
        ("off=[(m-f,23)];on=(m-s,10);tz=pt",
         {'off': [{'days': [0, 1, 2, 3, 4], 'hour': 23}],
          'on': [{'days': [0, 1, 2, 3, 4, 5], 'hour': 10}],
          'tz': 'pt'}),
        ('off=(m-f,19);on=(m-f,7);tz=pst',
         {'off': [{'days': [0, 1, 2, 3, 4], 'hour': 19}],
          'on': [{'days': [0, 1, 2, 3, 4], 'hour': 7}],
          'tz': 'pst'}),
        # wrap around days (saturday, sunday, monday)
        ('on=[(s-m,10)];off=(s-m,19)',
         {'on': [{'days': [5, 6, 0], 'hour': 10}],
          'off': [{'days': [5, 6, 0], 'hour': 19}],
          'tz': 'et'}),
        # multiple single days specified
        ('on=[(m,9),(t,10),(w,7)];off=(m-u,19)',
         {'on': [{'days': [0], 'hour': 9},
                 {'days': [1], 'hour': 10},
                 {'days': [2], 'hour': 7}],
          'off': [{'days': [0, 1, 2, 3, 4, 5, 6], 'hour': 19}],
          'tz': 'et'}),
        # using brackets also works, if only single time set
        ('off=[m-f,20];on=[m-f,5];tz=est',
         {'on': [{'days': [0, 1, 2, 3, 4], 'hour': 5}],
          'off': [{'days': [0, 1, 2, 3, 4], 'hour': 20}],
          'tz': 'est'}),
        # same string, exercise cache lookup.
        ('off=[m-f,20];on=[m-f,5];tz=est',
         {'on': [{'days': [0, 1, 2, 3, 4], 'hour': 5}],
          'off': [{'days': [0, 1, 2, 3, 4], 'hour': 20}],
          'tz': 'est'}),

        ################
        # Invalid Cases
        ('', None),
        # invalid day
        ('off=(1-2,12);on=(m-f,10);tz=est', None),
        # invalid hour
        ('off=(m-f,a);on=(m-f,10);tz=est', None),
        ('off=(m-f,99);on=(m-f,7);tz=pst', None),
        # invalid day
        ('off=(x-f,10);on=(m-f,10);tz=est', None),
        # no hour specified for on
        ('off=(m-f);on=(m-f,10);tz=est', None),
        # invalid day spec
        ('off=(m-t-f,12);on=(m-f,10);tz=est', None),
        # random extra
        ('off=(m-f,5);zebra=blue,on=(t-w,5)', None),
        ('off=(m-f,5);zebra=blue;on=(t-w,5)', None),
        # random extra again
        ('off=(m-f,5);zebrablue,on=(t-w,5)', None),
        ('bar;off=(m-f,5);zebrablue,on=(t-w,5)', None),
    ]

    def test_schedule_parser(self):
        self.maxDiff = None
        parser = ScheduleParser({'tz': 'et'})
        for value, expected in self.table:
            self.assertEqual(parser.parse(value), expected)



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from common import BaseTest

class CloudFront(BaseTest):

    def test_distribution_metric_filter(self):
        factory = self.replay_flight_data('test_distribution_metric_filter')
        p = self.load_policy({
            'name': 'requests-filter',
            'resource': 'distribution',
            'filters': [{
                'type': 'metrics',
                'name': 'Requests',
                'value': 3,
                'op': 'ge'
            }]
        }, session_factory=factory)
        resources = p.run()
        self.assertEqual(
            resources[0]['DomainName'], 'd1k7b41j4nj6pa.cloudfront.net')



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from common import BaseTest


class RDSClusterTest(BaseTest):

    def test_rdscluster_security_group(self):
        session_factory = self.replay_flight_data('test_rdscluster_sg_filter')
        p = self.load_policy({
            'name': 'rdscluster-sg',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'security-group',
                 'key': 'GroupName',
                 'value': 'default'}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['DatabaseName'], 'devtest')

    def test_rdscluster_subnet(self):
        session_factory = self.replay_flight_data('test_rdscluster_subnet')
        p = self.load_policy({
            'name': 'rdscluster-sub',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'subnet',
                 'key': 'MapPublicIpOnLaunch',
                 'value': True}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['DatabaseName'], 'devtest')

    def test_rdscluster_simple(self):
        session_factory = self.replay_flight_data('test_rdscluster_simple')
        p = self.load_policy({
            'name': 'rdscluster-simple',
            'resource': 'rds-cluster'},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 2)

    def test_rdscluster_simple_filter(self):
        session_factory = self.replay_flight_data('test_rdscluster_simple')
        p = self.load_policy({
            'name': 'rdscluster-simple-filter',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'value',
                 'key': 'DBClusterIdentifier',
                 'value': 'bbb'}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_rdscluster_delete(self):
        session_factory = self.replay_flight_data('test_rdscluster_delete')
        p = self.load_policy({
            'name': 'rdscluster-delete',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'value',
                 'key': 'DBClusterIdentifier',
                 'value': 'bbb'}],
            'actions': [
                {'type': 'delete',
                 'delete-instances': False}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_rdscluster_delete_with_instances(self):
        session_factory = self.replay_flight_data('test_rdscluster_delete_with_instances')
        p = self.load_policy({
            'name': 'rdscluster-delete',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'value',
                 'key': 'DBClusterIdentifier',
                 'value': 'bbb'}],
            'actions': [
                {'type': 'delete',
                 'delete-instances': True}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_rdscluster_retention(self):
        session_factory = self.replay_flight_data('test_rdscluster_retention')
        p = self.load_policy({
            'name': 'rdscluster-delete',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'value',
                 'key': 'DBClusterIdentifier',
                 'value': 'bbb'}],
            'actions': [{'type': 'retention', 'days': 21}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_rdscluster_snapshot(self):
        session_factory = self.replay_flight_data('test_rdscluster_snapshot')
        p = self.load_policy({
            'name': 'rdscluster-snapshot',
            'resource': 'rds-cluster',
            'filters': [
                {'type': 'value',
                 'key': 'DBClusterIdentifier',
                 'value': 'bbb'}],
            'actions': [{'type': 'snapshot'}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)


class RDSClusterSnapshotTest(BaseTest):

    def test_rdscluster_snapshot_simple(self):
        session_factory = self.replay_flight_data(
            'test_rdscluster_snapshot_simple')
        p = self.load_policy({
            'name': 'rdscluster-snapshot-simple',
            'resource': 'rds-cluster-snapshot'},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 2)

    def test_rdscluster_snapshot_simple_filter(self):
        session_factory = self.replay_flight_data(
            'test_rdscluster_snapshot_simple')
        p = self.load_policy({
            'name': 'rdscluster-snapshot-simple-filter',
            'resource': 'rds-cluster-snapshot',
            'filters': [
                {'type': 'value',
                 'key': 'StorageEncrypted',
                 'value': False}]},
            session_factory=session_factory)
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_rdscluster_snapshot_age_filter(self):
        factory = self.replay_flight_data('test_rdscluster_snapshot_simple')
        p = self.load_policy({
            'name': 'rdscluster-snapshot-age-filter',
            'resource': 'rds-cluster-snapshot',
            'filters': [{'type': 'age', 'days': 7}]},
            session_factory=factory)
        resources = p.run()
        self.assertEqual(len(resources), 2)

    def test_rdscluster_snapshot_trim(self):
        factory = self.replay_flight_data('test_rdscluster_snapshot_delete')
        p = self.load_policy({
            'name': 'rdscluster-snapshot-trim',
            'resource': 'rds-cluster-snapshot',
            'actions': ['delete']},
            session_factory=factory)
        resources = p.run()
        self.assertEqual(len(resources), 2)



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .core import (
    ANNOTATION_KEY,
    FilterValidationError,
    OPERATORS,
    FilterRegistry,
    Filter,
    Or,
    And,
    ValueFilter,
    AgeFilter,
    EventFilter)
from .iamaccess import CrossAccountAccessFilter
from .metrics import MetricsFilter
from .vpc import DefaultVpcBase



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from c7n import cache
from c7n.executor import ThreadPoolExecutor
from c7n.registry import PluginRegistry
from c7n.utils import dumps


resources = PluginRegistry('resources')


class ResourceManager(object):

    filter_registry = None
    action_registry = None
    executor_factory = ThreadPoolExecutor
    retry = None

    def __init__(self, ctx, data):
        self.ctx = ctx
        self.session_factory = ctx.session_factory
        self.config = ctx.options
        self.data = data
        self.log_dir = ctx.log_dir
        self._cache = cache.factory(self.ctx.options)
        self.log = logging.getLogger('custodian.resources.%s' % (
            self.__class__.__name__.lower()))

        if self.filter_registry:
            self.filters = self.filter_registry.parse(
                self.data.get('filters', []), self)
        if self.action_registry:
            self.actions = self.action_registry.parse(
                self.data.get('actions', []), self)

    def format_json(self, resources, fh):
        return dumps(resources, fh, indent=2)

    def match_ids(self, ids):
        """return ids that match this resource type's id format."""
        return ids

    @classmethod
    def get_permissions(cls):
        return ()

    def get_resources(self, resource_ids):
        """Retrieve a set of resources by id."""
        return []

    def resources(self):
        raise NotImplementedError("")

    def get_resource_manager(self, resource_type, data=None):
        klass = resources.get(resource_type)
        if klass is None:
            raise ValueError(resource_type)
        # if we're already querying via config carry it forward
        if not data and self.source_type == 'config' and getattr(
                klass.get_model(), 'config_type', None):
            return klass(self.ctx, {'source': self.config_type})
        return klass(self.ctx, data or {})

    def filter_resources(self, resources, event=None):
        original = len(resources)
        if event and event.get('debug', False):
            self.log.info(
                "Filtering resources with %s", self.filters)
        for f in self.filters:
            if not resources:
                break
            rcount = len(resources)
            resources = f.process(resources, event)
            if event and event.get('debug', False):
                self.log.debug(
                    "applied filter %s %d->%d", f, rcount, len(resources))
        self.log.debug("Filtered from %d to %d %s" % (
            original, len(resources), self.__class__.__name__.lower()))
        return resources

    def get_model(self):
        """Returns the resource meta-model.
        """
        return self.query.resolve(self.resource_type)



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from common import BaseTest
from c7n.resources.vpc import SecurityGroupDiff, SecurityGroupPatch


class SGDiffLibTest(BaseTest):

    def test_sg_diff_pitr(self):
        factory = self.replay_flight_data('test_sg_config_diff')
        p = self.load_policy({
            'name': 'sg-differ',
            'resource': 'security-group',
            'filters': [
                {'GroupId': 'sg-a38ed1de'},
                {'type': 'diff',
                 'selector': 'date',
                 'selector_value': '2016/12/11 17:25Z'}],
        }, session_factory=factory)

        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.maxDiff = None
        self.assertEqual(resources[0]['c7n:diff'],
            {'egress': {
                'added': [{u'IpProtocol': u'-1',
                           u'IpRanges': [{u'CidrIp': u'0.0.0.0/0'}],
                           u'Ipv6Ranges': [],
                           u'PrefixListIds': [],
                           u'UserIdGroupPairs': [{u'GroupId': u'sg-a08ed1dd',
                                                  u'UserId': u'644160558196'}]}],
                'removed': [{
                             u'IpProtocol': u'-1',
                             u'IpRanges': [{u'CidrIp': u'0.0.0.0/0'}],
                             u'Ipv6Ranges': [],
                             u'PrefixListIds': [],
                             u'UserIdGroupPairs': []}]},
             'ingress': {
                 'added': [{u'FromPort': 8485,
                            u'IpProtocol': u'tcp',
                            u'IpRanges': [],
                            u'Ipv6Ranges': [],
                            u'PrefixListIds': [],
                            u'ToPort': 8485,
                            u'UserIdGroupPairs': [{u'GroupId': u'sg-a38ed1de',
                                                   u'UserId': u'644160558196'}]},
                           {u'FromPort': 22,
                            u'IpProtocol': u'tcp',
                            u'IpRanges': [{u'CidrIp': u'10.0.0.0/24'}],
                            u'Ipv6Ranges': [],
                            u'PrefixListIds': [],
                            u'ToPort': 22,
                            u'UserIdGroupPairs': []}]},
             'tags': {'added': {u'Scope': u'account'}}})

    def test_sg_patch_pitr(self):
        factory = self.replay_flight_data('test_sg_config_patch_pitr')
        p = self.load_policy({
            'name': 'sg-differ',
            'resource': 'security-group',
            'filters': [
                {'GroupId': 'sg-a38ed1de'},
                {'type': 'diff',
                 'selector': 'date',
                 'selector_value': '2016/12/11 17:25Z'}],
            'actions': ['patch']
        }, session_factory=factory)

        resources = p.run()
        self.assertEqual(len(resources), 1)

        current_resource = factory().client('ec2').describe_security_groups(
            GroupIds=['sg-a38ed1de'])['SecurityGroups'][0]

        self.maxDiff = None
        self.assertEqual(
            current_resource,
            resources[0]['c7n:previous-revision']['resource'])
                
    def test_sg_diff_patch(self):
        factory = self.replay_flight_data(
            'test_security_group_revisions_delta')
        client = factory().client('ec2')
        vpc_id = client.create_vpc(CidrBlock="10.42.0.0/16")['Vpc']['VpcId']
        self.addCleanup(client.delete_vpc, VpcId=vpc_id)
        sg_id = client.create_security_group(
            GroupName="allow-access",
            VpcId=vpc_id,
            Description="inbound access")['GroupId']
        self.addCleanup(client.delete_security_group, GroupId=sg_id)

        client.create_tags(
            Resources=[sg_id],
            Tags=[
                {'Key': 'NetworkLocation', 'Value': 'DMZ'},
                {'Key': 'App', 'Value': 'blue-moon'}
            ])
        client.authorize_security_group_ingress(
            GroupId=sg_id,
            IpPermissions=[
                {'IpProtocol': 'tcp',
                 'FromPort': 443,
                 'ToPort': 443,
                 'IpRanges': [{'CidrIp': '10.42.1.0/24'}]},
                {'IpProtocol': 'tcp',
                 'FromPort': 8080,
                 'ToPort': 8080,
                 'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}
                ])

        s1 = client.describe_security_groups(GroupIds=[sg_id])[
            'SecurityGroups'][0]

        # Modify state
        client.create_tags(
            Resources=[sg_id],
            Tags=[
                {'Key': 'App', 'Value': 'red-moon'},
                {'Key': 'Stage', 'Value': 'production'}])
        client.revoke_security_group_ingress(
            GroupId=sg_id,
            IpPermissions=[
                {'IpProtocol': 'tcp',
                 'FromPort': 8080,
                 'ToPort': 8080,
                 'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}])
        client.authorize_security_group_ingress(
            GroupId=sg_id,
            IpPermissions=[
                {'IpProtocol': 'tcp',
                 'FromPort': 80,
                 'ToPort': 80,
                 'IpRanges': [{'CidrIp': '0.0.0.0/0'}]},
                ])
        s2 = client.describe_security_groups(GroupIds=[sg_id])[
            'SecurityGroups'][0]

        # Apply reverse delta
        self.maxDiff = None
        self.assertEqual(
            {'ingress': {'added': [{u'FromPort': 80,
                                    u'IpProtocol': 'tcp',
                                    u'IpRanges': [{u'CidrIp': '0.0.0.0/0'}],
                                    u'Ipv6Ranges': [],
                                    u'PrefixListIds': [],
                                    u'ToPort': 80,
                                    u'UserIdGroupPairs': []}],
                         'removed': [{u'FromPort': 8080,
                                      u'IpProtocol': 'tcp',
                                      u'IpRanges': [{u'CidrIp': '0.0.0.0/0'}],
                                      u'Ipv6Ranges': [],
                                      u'PrefixListIds': [],
                                      u'ToPort': 8080,
                                      u'UserIdGroupPairs': []}]},
             'tags': {'added': {'Stage': 'production'},
                      'updated': {'App': 'red-moon'}}},
            SecurityGroupDiff().diff(s1, s2))

        SecurityGroupPatch().apply_delta(
            client, s2,
            SecurityGroupDiff().diff(s2, s1))

        # Compare to origin
        s3 = client.describe_security_groups(GroupIds=[sg_id])[
            'SecurityGroups'][0]

        self.assertEqual(s1, s3)



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Cloud Custodian Lambda Provisioning Support

docs/lambda.rst
"""

import abc
import base64
import imp
import hashlib
import json
import logging
import os
import StringIO
import sys
import time
import tempfile
import zipfile

from boto3.s3.transfer import S3Transfer, TransferConfig
from botocore.exceptions import ClientError

from concurrent.futures import ThreadPoolExecutor

# Static event mapping to help simplify cwe rules creation
from c7n.cwe import CloudWatchEvents
from c7n.logs_support import _timestamp_from_string
from c7n.utils import parse_s3, local_session


log = logging.getLogger('custodian.lambda')
RUNTIME = 'python{}.{}'.format(
    sys.version_info.major,
    sys.version_info.minor,
)


class PythonPackageArchive(object):
    """Creates a zip file for python lambda functions.

    :param tuple modules: the Python modules to add to the archive

    Amazon doesn't give us straightforward docs here, only `an example
    <http://docs.aws.amazon.com/lambda/latest/dg/with-s3-example-deployment-pkg.html#with-s3-example-deployment-pkg-python>`_,
    from which we can infer that they simply unzip the file into a directory on
    ``sys.path``. So what we do is locate all of the ``modules`` specified, and
    add all of the ``.py`` files we find for these modules to a zip file.

    In addition to the modules specified during instantiation, you can add
    arbitrary additional files to the archive using :py:func:`add_file` and
    :py:func:`add_contents`. For example, since we only add ``*.py`` files for
    you, you'll need to manually add files for any compiled extension modules
    that your Lambda requires.

    """

    def __init__(self, *modules):
        self._temp_archive_file = tempfile.NamedTemporaryFile()
        self._zip_file = zipfile.ZipFile(
            self._temp_archive_file, mode='w',
            compression=zipfile.ZIP_DEFLATED)
        self._closed = False
        self.add_modules(*modules)

    @property
    def path(self):
        return self._temp_archive_file.name

    @property
    def size(self):
        if not self._closed:
            raise ValueError("Archive not closed, size not accurate")
        return os.stat(self._temp_archive_file.name).st_size

    def add_modules(self, *modules):
        """Add the named Python modules to the archive. For consistency's sake
        we only add ``*.py`` files, not ``*.pyc``. We also don't add other
        files, including compiled modules. You'll have to add such files
        manually using :py:meth:`add_file`.
        """
        for module in modules:
            path = imp.find_module(module)[1]
            if os.path.isfile(path):
                if not path.endswith('.py'):
                    raise ValueError('We need a *.py source file instead of ' + path)
                self.add_file(path)
            elif os.path.isdir(path):
                for root, dirs, files in os.walk(path):
                    arc_prefix = os.path.relpath(root, os.path.dirname(path))
                    for f in files:
                        if not f.endswith('.py'):
                            continue
                        f_path = os.path.join(root, f)
                        dest_path = os.path.join(arc_prefix, f)
                        self.add_file(f_path, dest_path)

    def add_file(self, src, dest=None):
        """Add the file at ``src`` to the archive.

        If ``dest`` is ``None`` then it is added under just the original
        filename. So ``add_file('foo/bar.txt')`` ends up at ``bar.txt`` in the
        archive, while ``add_file('bar.txt', 'foo/bar.txt')`` ends up at
        ``foo/bar.txt``.

        """
        dest = dest or os.path.basename(src)
        with open(src, 'rb') as fp:
            contents = fp.read()
        self.add_contents(dest, contents)

    def add_py_file(self, src, dest=None):
        """This is a special case of :py:meth:`add_file` that helps for adding
        a ``py`` when a ``pyc`` may be present as well. So for example, if
        ``__file__`` is ``foo.pyc`` and you do:

        .. code-block:: python

          archive.add_py_file(__file__)

        then this method will add ``foo.py`` instead if it exists, and raise
        ``IOError`` if it doesn't.

        """
        src = src[:-1] if src.endswith('.pyc') else src
        self.add_file(src, dest)

    def add_contents(self, dest, contents):
        """Add file contents to the archive under ``dest``.

        If ``dest`` is a path, it will be added compressed and world-readable
        (user-writeable). You may also pass a :py:class:`~zipfile.ZipInfo` for
        custom behavior.

        """
        assert not self._closed, "Archive closed"
        if not isinstance(dest, zipfile.ZipInfo):
            dest = zinfo(dest)  # see for some caveats
        self._zip_file.writestr(dest, contents)

    def close(self):
        """Close the zip file.

        Note underlying tempfile is removed when archive is garbage collected.
        """
        self._closed = True
        self._zip_file.close()
        log.debug(
            "Created custodian lambda archive size: %0.2fmb",
            (os.path.getsize(self._temp_archive_file.name) / (
                1024.0 * 1024.0)))
        return self

    def remove(self):
        """Dispose of the temp file for garbage collection."""
        if self._temp_archive_file:
            self._temp_archive_file = None

    def get_checksum(self):
        """Return the b64 encoded sha256 checksum of the archive."""
        assert self._closed, "Archive not closed"
        with open(self._temp_archive_file.name) as fh:
            return base64.b64encode(checksum(fh, hashlib.sha256()))

    def get_bytes(self):
        """Return the entire zip file as a byte string. """
        assert self._closed, "Archive not closed"
        return open(self._temp_archive_file.name, 'rb').read()

    def get_reader(self):
        """Return a read-only :py:class:`~zipfile.ZipFile`."""
        assert self._closed, "Archive not closed"
        io = StringIO.StringIO(self.get_bytes())
        return zipfile.ZipFile(io, mode='r')

    def get_filenames(self):
        """Return a list of filenames in the archive."""
        return [n.filename for n in self.get_reader().filelist]


def checksum(fh, hasher, blocksize=65536):
    buf = fh.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fh.read(blocksize)
    return hasher.digest()


def custodian_archive():
    """Create a lambda code archive for running custodian."""
    return PythonPackageArchive('c7n', 'pkg_resources', 'ipaddress')


class LambdaManager(object):
    """ Provides CRUD operations around lambda functions
    """

    def __init__(self, session_factory, s3_asset_path=None):
        self.session_factory = session_factory
        self.client = self.session_factory().client('lambda')
        self.s3_asset_path = s3_asset_path

    def list_functions(self, prefix=None):
        p = self.client.get_paginator('list_functions')
        for rp in p.paginate():
            for f in rp.get('Functions', []):
                if not prefix:
                    yield f
                elif f['FunctionName'].startswith(prefix):
                    yield f

    def publish(self, func, alias=None, role=None, s3_uri=None):
        result, changed = self._create_or_update(
            func, role, s3_uri, qualifier=alias)
        func.arn = result['FunctionArn']
        if alias and changed:
            func.alias = self.publish_alias(result, alias)
        elif alias:
            func.alias = "%s:%s" % (func.arn, alias)
        else:
            func.alias = func.arn

        for e in func.get_events(self.session_factory):
            if e.add(func):
                log.debug(
                    "Added event source: %s to function: %s",
                    e, func.alias)
        return result

    def remove(self, func, alias=None):
        log.info("Removing lambda function %s", func.name)
        for e in func.get_events(self.session_factory):
            e.remove(func)
        try:
            self.client.delete_function(FunctionName=func.name)
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceNotFoundException':
                raise

    def metrics(self, funcs, start, end, period=5 * 60):

        def func_metrics(f):
            metrics = local_session(self.session_factory).client('cloudwatch')
            values = {}
            for m in ('Errors', 'Invocations', 'Durations', 'Throttles'):
                values[m] = metrics.get_metric_statistics(
                    Namespace="AWS/Lambda",
                    Dimensions=[{
                        'Name': 'FunctionName',
                        'Value': (
                            isinstance(f, dict) and f['FunctionName'] or f.name)}],
                    Statistics=["Sum"],
                    StartTime=start,
                    EndTime=end,
                    Period=period,
                    MetricName=m)['Datapoints']
            return values

        with ThreadPoolExecutor(max_workers=3) as w:
            results = list(w.map(func_metrics, funcs))
            for m, f in zip(results, funcs):
                if isinstance(f, dict):
                    f['Metrics'] = m
        return results

    def logs(self, func, start, end):
        logs = self.session_factory().client('logs')
        group_name = "/aws/lambda/%s" % func.name
        log.info("Fetching logs from group: %s" % group_name)
        try:
            logs.describe_log_groups(
                logGroupNamePrefix=group_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                return
            raise
        try:
            log_streams = logs.describe_log_streams(
                logGroupName=group_name,
                orderBy="LastEventTime", limit=3, descending=True)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                return
            raise
        start = _timestamp_from_string(start)
        end = _timestamp_from_string(end)
        for s in reversed(log_streams['logStreams']):
            result = logs.get_log_events(
                logGroupName=group_name,
                logStreamName=s['logStreamName'],
                startTime=start,
                endTime=end,
            )
            for e in result['events']:
                yield e

    @staticmethod
    def delta_function(lambda_func, func, role):
        conf = func.get_config()
        # TODO feels a little wierd
        conf['Role'] = role
        for k in conf:
            if conf[k] != lambda_func['Configuration'][k]:
                return True

    def _create_or_update(self, func, role=None, s3_uri=None, qualifier=None):
        role = func.role or role
        assert role, "Lambda function role must be specified"
        archive = func.get_archive()
        lfunc = self.get(func.name, qualifier)

        if s3_uri:
            # TODO: support versioned buckets
            bucket, key = self._upload_func(s3_uri, func, archive)
            code_ref = {'S3Bucket': bucket, 'S3Key': key}
        else:
            code_ref = {'ZipFile': archive.get_bytes()}

        changed = False
        if lfunc:
            result = lfunc['Configuration']
            if archive.get_checksum() != lfunc['Configuration']['CodeSha256']:
                log.debug("Updating function %s code", func.name)
                params = dict(FunctionName=func.name, Publish=True)
                params.update(code_ref)
                result = self.client.update_function_code(**params)
                changed = True
            # TODO/Consider also set publish above to false, and publish
            # after configuration change?
            if self.delta_function(lfunc, func, role):
                log.debug("Updating function: %s config" % func.name)
                params = func.get_config()
                del params['Runtime']
                params['Role'] = role
                result = self.client.update_function_configuration(**params)
                changed = True
        else:
            log.info('Publishing custodian policy lambda function %s', func.name)
            params = func.get_config()
            params.update({'Publish': True, 'Code': code_ref, 'Role': role})
            result = self.client.create_function(**params)
            changed = True

        return result, changed

    def _upload_func(self, s3_uri, func, archive):
        _, bucket, key_prefix = parse_s3(s3_uri)
        key = "%s/%s" % (key_prefix, func.name)
        transfer = S3Transfer(
            self.session_factory().client('s3'),
            config=TransferConfig(
                multipart_threshold=1024 * 1024 * 4))
        transfer.upload_file(
            archive.path,
            bucket=bucket,
            key=key,
            extra_args={
                'ServerSideEncryption': 'AES256'})
        return bucket, key

    def publish_alias(self, func_data, alias):
        """Create or update an alias for the given function.
        """
        if not alias:
            return func_data['FunctionArn']
        func_name = func_data['FunctionName']
        func_version = func_data['Version']

        exists = resource_exists(
            self.client.get_alias, FunctionName=func_name, Name=alias)

        if not exists:
            log.debug("Publishing custodian lambda alias %s", alias)
            alias_result = self.client.create_alias(
                FunctionName=func_name,
                Name=alias,
                FunctionVersion=func_version)
        else:
            if (exists['FunctionVersion'] == func_version and
                    exists['Name'] == alias):
                return exists['AliasArn']
            log.debug('Updating custodian lambda alias %s', alias)
            alias_result = self.client.update_alias(
                FunctionName=func_name,
                Name=alias,
                FunctionVersion=func_version)
        return alias_result['AliasArn']

    def get(self, func_name, qualifier=None):
        params = {'FunctionName': func_name}
        if qualifier:
            params['Qualifier'] = qualifier
        return resource_exists(
            self.client.get_function, **params)


def resource_exists(op, NotFound="ResourceNotFoundException", *args, **kw):
    try:
        return op(*args, **kw)
    except ClientError, e:
        if e.response['Error']['Code'] == NotFound:
            return False
        raise


class AbstractLambdaFunction:
    """Abstract base class for lambda functions."""
    __metaclass__ = abc.ABCMeta

    alias = None

    @abc.abstractproperty
    def name(self):
        """Name for the lambda function"""

    @abc.abstractproperty
    def runtime(self):
        """ """

    @abc.abstractproperty
    def description(self):
        """ """

    @abc.abstractproperty
    def handler(self):
        """ """

    @abc.abstractproperty
    def memory_size(self):
        """ """

    @abc.abstractproperty
    def timeout(self):
        """ """

    @abc.abstractproperty
    def role(self):
        """ """

    @abc.abstractproperty
    def subnets(self):
        """ """

    @abc.abstractproperty
    def security_groups(self):
        """ """

    @abc.abstractmethod
    def get_events(self, session_factory):
        """event sources that should be bound to this lambda."""

    @abc.abstractmethod
    def get_archive(self):
        """Return the lambda distribution archive object."""

    def get_config(self):
        conf = {
            'FunctionName': self.name,
            'MemorySize': self.memory_size,
            'Role': self.role,
            'Description': self.description,
            'Runtime': self.runtime,
            'Handler': self.handler,
            'Timeout': self.timeout}
        if self.subnets and self.security_groups:
            conf['VpcConfig'] = {
                'SubnetIds': self.subnets,
                'SecurityGroupIds': self.security_groups}
        return conf


class LambdaFunction(AbstractLambdaFunction):

    def __init__(self, func_data, archive):
        self.func_data = func_data
        required = set((
            'name', 'handler', 'memory_size',
            'timeout', 'role', 'runtime',
            'description'))
        missing = required.difference(func_data)
        if missing:
            raise ValueError("Missing required keys %s" % " ".join(missing))
        self.archive = archive

    @property
    def name(self):
        return self.func_data['name']

    @property
    def description(self):
        return self.func_data['description']

    @property
    def handler(self):
        return self.func_data['handler']

    @property
    def memory_size(self):
        return self.func_data['memory_size']

    @property
    def timeout(self):
        return self.func_data['timeout']

    @property
    def runtime(self):
        return self.func_data['runtime']

    @property
    def role(self):
        return self.func_data['role']

    @property
    def security_groups(self):
        return self.func_data.get('security_groups', None)

    @property
    def subnets(self):
        return self.func_data.get('subnets', None)

    def get_events(self, session_factory):
        return self.func_data.get('events', ())

    def get_archive(self):
        return self.archive


PolicyHandlerTemplate = """\
from c7n import handler

def run(event, context):
    return handler.dispatch_event(event, context)

"""


class PolicyLambda(AbstractLambdaFunction):
    """Wraps a custodian policy to turn it into lambda function.
    """
    handler = "custodian_policy.run"
    runtime = "python2.7"

    def __init__(self, policy):
        self.policy = policy
        self.archive = custodian_archive()

    @property
    def name(self):
        return "custodian-%s" % self.policy.name

    @property
    def description(self):
        return self.policy.data.get(
            'description', 'cloud-custodian lambda policy')

    @property
    def role(self):
        return self.policy.data['mode'].get('role', '')

    @property
    def memory_size(self):
        return self.policy.data['mode'].get('memory', 512)

    @property
    def timeout(self):
        return self.policy.data['mode'].get('timeout', 60)

    @property
    def security_groups(self):
        return None

    @property
    def subnets(self):
        return None

    def get_events(self, session_factory):
        events = []
        if self.policy.data['mode']['type'] == 'config-rule':
            events.append(
                ConfigRule(self.policy.data['mode'], session_factory))
        else:
            events.append(
                CloudWatchEventSource(
                    self.policy.data['mode'], session_factory))
        return events

    def get_archive(self):
        self.archive.add_contents(
            'config.json', json.dumps(
                {'policies': [self.policy.data]}, indent=2))
        self.archive.add_contents('custodian_policy.py', PolicyHandlerTemplate)
        self.archive.close()
        return self.archive


def zinfo(fname):
    """Amazon lambda exec environment setup can break itself
    if zip files aren't constructed a particular way.

    ie. It respects file perm attributes from the zip including
    those that prevent lambda from working. Namely lambda
    extracts code as one user, and executes code as a different
    user. Without permissions for the executing user to read
    the file the lambda function is broken.

    Python's default zipfile.writestr does a 0600 perm which
    we modify here as a workaround.
    """
    info = zipfile.ZipInfo(fname)
    # Grant other users permissions to read
    # http://unix.stackexchange.com/questions/14705/
    info.external_attr = 0o644 << 16L
    return info


class CloudWatchEventSource(object):
    """Subscribe a lambda to cloud watch events.

    Cloud watch events supports a number of different event
    sources, from periodic timers with cron syntax, to
    real time instance state notifications, cloud trail
    events, and realtime asg membership changes.

    Event Pattern for Instance State

    .. code-block:: json

       {
         "source": ["aws.ec2"],
         "detail-type": ["EC2 Instance State-change Notification"],
         "detail": { "state": ["pending"]}
       }

    Event Pattern for Cloud Trail API

    .. code-block:: json

       {
         "detail-type": ["AWS API Call via CloudTrail"],
         "detail": {
            "eventSource": ["s3.amazonaws.com"],
            "eventName": ["CreateBucket", "DeleteBucket"]
         }
       }
    """
    ASG_EVENT_MAPPING = {
        'launch-success': 'EC2 Instance Launch Successful',
        'launch-failure': 'EC2 Instance Launch Unsuccessful',
        'terminate-success': 'EC2 Instance Terminate Successful',
        'terminate-failure': 'EC2 Instance Terminate Unsuccessful'}

    def __init__(self, data, session_factory, prefix="custodian-"):
        self.session_factory = session_factory
        self.session = session_factory()
        self.client = self.session.client('events')
        self.data = data
        self.prefix = prefix

    def _make_notification_id(self, function_name):
        if not function_name.startswith(self.prefix):
            return "%s%s" % (self.prefix, function_name)
        return function_name

    def get(self, rule_name):
        return resource_exists(
            self.client.describe_rule,
            Name=self._make_notification_id(rule_name))

    @staticmethod
    def delta(src, tgt):
        """Given two cwe rules determine if the configuration is the same.

        Name is already implied.
        """
        for k in ['State', 'EventPattern', 'ScheduleExpression']:
            if src.get(k) != tgt.get(k):
                return True
        return False

    def __repr__(self):
        return "<CWEvent Type:%s Events:%s>" % (
            self.data.get('type'),
            ', '.join(map(str, self.data.get('events', []))))

    def resolve_cloudtrail_payload(self, payload):
        sources = self.data.get('sources', [])
        events = []
        for e in self.data.get('events'):
            if not isinstance(e, dict):
                events.append(e)
                event_info = CloudWatchEvents.get(e)
                if event_info is None:
                    continue
            else:
                event_info = e
                events.append(e['event'])
            sources.append(event_info['source'])

        payload['detail'] = {
            'eventSource': list(set(sources)),
            'eventName': events}

    def render_event_pattern(self):
        event_type = self.data.get('type')
        payload = {}
        if event_type == 'cloudtrail':
            payload['detail-type'] = ['AWS API Call via CloudTrail']
            self.resolve_cloudtrail_payload(payload)

        elif event_type == "ec2-instance-state":
            payload['source'] = ['aws.ec2']
            payload['detail-type'] = [
                "EC2 Instance State-change Notification"]
            # Technically could let empty be all events, but likely misconfig
            payload['detail'] = {"state": self.data.get('events', [])}
        elif event_type == "asg-instance-state":
            payload['source'] = ['aws.autoscaling']
            events = []
            for e in self.data.get('events', []):
                events.append(self.ASG_EVENT_MAPPING.get(e, e))
            payload['detail-type'] = events
        elif event_type == 'periodic':
            pass
        else:
            raise ValueError(
                "Unknown lambda event source type: %s" % event_type)
        if not payload:
            return None
        return json.dumps(payload)

    def add(self, func):
        params = dict(
            Name=func.name, Description=func.description, State='ENABLED')

        pattern = self.render_event_pattern()
        if pattern:
            params['EventPattern'] = pattern
        schedule = self.data.get('schedule')
        if schedule:
            params['ScheduleExpression'] = schedule

        rule = self.get(func.name)

        if rule and self.delta(rule, params):
            log.debug("Updating cwe rule for %s" % self)
            response = self.client.put_rule(**params)
        elif not rule:
            log.debug("Creating cwe rule for %s" % (self))
            response = self.client.put_rule(**params)
        else:
            response = {'RuleArn': rule['Arn']}

        try:
            self.session.client('lambda').add_permission(
                FunctionName=func.name,
                StatementId=func.name,
                SourceArn=response['RuleArn'],
                Action='lambda:InvokeFunction',
                Principal='events.amazonaws.com')
            log.debug('Added lambda invoke cwe rule permission')
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceConflictException':
                raise

        # Add Targets
        found = False
        response = self.client.list_targets_by_rule(Rule=func.name)
        # CWE seems to be quite picky about function arns (no aliases/versions)
        func_arn = func.arn

        if func_arn.count(':') > 6:
            func_arn, version = func_arn.rsplit(':', 1)
        for t in response['Targets']:
            if func_arn == t['Arn']:
                found = True

        if found:
            return

        log.debug('Creating cwe rule target for %s on func:%s' % (
            self, func_arn))

        self.client.put_targets(
            Rule=func.name, Targets=[{"Id": func.name, "Arn": func_arn}])

        return True

    def update(self, func):
        self.add(func)

    def pause(self, func):
        try:
            self.client.disable_rule(Name=func.name)
        except:
            pass

    def resume(self, func):
        try:
            self.client.enable_rule(Name=func.name)
        except:
            pass

    def remove(self, func):
        if self.get(func.name):
            try:
                targets = self.client.list_targets_by_rule(
                    Rule=func.name)['Targets']
                self.client.remove_targets(
                    Rule=func.name,
                    Ids=[t['Id'] for t in targets])
            except ClientError as e:
                log.warning(
                    "Could not remove targets for rule %s error: %s",
                    func.name, e)
            self.client.delete_rule(Name=func.name)


class BucketLambdaNotification(object):
    """ Subscribe a lambda to bucket notifications directly. """

    def __init__(self, data, session_factory, bucket):
        self.data = data
        self.session_factory = session_factory
        self.session = session_factory()
        self.bucket = bucket

    def delta(self, src, tgt):
        for k in ['Id', 'LambdaFunctionArn', 'Events', 'Filters']:
            if src.get(k) != tgt.get(k):
                return True
        return False

    def _get_notifies(self, s3, func):
        notifies = s3.get_bucket_notification_configuration(
            Bucket=self.bucket['Name'])
        found = False
        for f in notifies.get('LambdaFunctionConfigurations', []):
            if f['Id'] != func.name:
                continue
            found = f
        return notifies, found

    def add(self, func):
        s3 = self.session.client('s3')
        notifies, found = self._get_notifies(s3, func)
        notifies.pop('ResponseMetadata', None)
        func_arn = func.arn
        if func_arn.rsplit(':', 1)[-1].isdigit():
            func_arn = func_arn.rsplit(':', 1)[0]
        n_params = {
            'Id': func.name,
            'LambdaFunctionArn': func_arn,
            'Events': self.data.get('events', ['s3:ObjectCreated:*'])}
        if self.data.get('filters'):
            n_params['Filters'] = {
                'Key': {'FilterRules': self.filters}}

        if found:
            if self.delta(found, n_params):
                notifies['LambdaFunctionConfigurations'].remove(found)
            else:
                log.info("Bucket lambda notification present")
                return

        lambda_client = self.session.client('lambda')
        params = dict(
            FunctionName=func.name,
            StatementId=self.bucket['Name'],
            Action='lambda:InvokeFunction',
            Principal='s3.amazonaws.com')
        if self.data.get('account_s3'):
            params['SourceAccount'] = self.data['account_s3']
            params['SourceArn'] = 'arn:aws:s3:::*'
        else:
            params['SourceArn'] = 'arn:aws:s3:::%' % self.bucket['Name']
        try:
            lambda_client.add_permission(**params)
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceConflictException':
                raise

        notifies.setdefault('LambdaFunctionConfigurations', []).append(n_params)
        s3.put_bucket_notification_configuration(
            Bucket=self.bucket['Name'], NotificationConfiguration=notifies)

        return True

    def remove(self, func):
        s3 = self.session.client('s3')
        notifies, found = self._get_notifies(s3, func)
        if not found:
            return

        lambda_client = self.session.client('lambda')
        try:
            response = lambda_client.remove_permission(
                FunctionName=func['FunctionName'],
                StatementId=self.bucket['Name'])
            log.debug("Removed lambda permission result: %s" % response)
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceNotFoundException':
                raise

        notifies['LambdaFunctionConfigurations'].remove(found)
        s3.put_bucket_notification_configuration(
            Bucket=self.bucket['Name'],
            NotificationConfiguration=notifies)


class CloudWatchLogSubscription(object):
    """ Subscribe a lambda to a log group[s]
    """

    iam_delay = 1.5

    def __init__(self, session_factory, log_groups, filter_pattern):
        self.log_groups = log_groups
        self.filter_pattern = filter_pattern
        self.session_factory = session_factory
        self.session = session_factory()
        self.client = self.session.client('logs')

    def add(self, func):
        lambda_client = self.session.client('lambda')
        for group in self.log_groups:
            log.info(
                "Creating subscription filter for %s" % group['logGroupName'])
            region = group['arn'].split(':', 4)[3]
            try:
                lambda_client.add_permission(
                    FunctionName=func.name,
                    StatementId=group['logGroupName'][1:].replace('/', '-'),
                    SourceArn=group['arn'],
                    Action='lambda:InvokeFunction',
                    Principal='logs.%s.amazonaws.com' % region)
                log.debug("Added lambda ipo nvoke log group permission")
                # iam eventual consistency and propagation
                time.sleep(self.iam_delay)
            except ClientError as e:
                if e.response['Error']['Code'] != 'ResourceConflictException':
                    raise
            # Consistent put semantics / ie no op if extant
            self.client.put_subscription_filter(
                logGroupName=group['logGroupName'],
                filterName=func.name,
                filterPattern=self.filter_pattern,
                destinationArn=func.alias or func.arn)

    def remove(self, func):
        lambda_client = self.session.client('lambda')
        for group in self.log_groups:
            try:
                response = lambda_client.remove_permission(
                    FunctionName=func.name,
                    StatementId=group['logGroupName'][1:].replace('/', '-'))
                log.debug("Removed lambda permission result: %s" % response)
            except ClientError as e:
                if e.response['Error']['Code'] != 'ResourceNotFoundException':
                    raise

            try:
                response = self.client.delete_subscription_filter(
                    logGroupName=group['logGroupName'], filterName=func.name)
                log.debug("Removed subscription filter from: %s",
                          group['logGroupName'])
            except ClientError as e:
                if e.response['Error']['Code'] != 'ResourceNotFoundException':
                    raise


class SNSSubscription(object):
    """ Subscribe a lambda to one or more SNS topics.
    """

    iam_delay = 1.5

    def __init__(self, session_factory, topic_arns):
        self.topic_arns = topic_arns
        self.session_factory = session_factory
        self.session = session_factory()
        self.client = self.session.client('sns')

    @staticmethod
    def _parse_arn(arn):
        parts = arn.split(':')
        region, topic_name = parts[3], parts[5]
        statement_id = 'sns-topic-' + topic_name
        return region, topic_name, statement_id

    def add(self, func):
        lambda_client = self.session.client('lambda')
        for arn in self.topic_arns:
            region, topic_name, statement_id = self._parse_arn(arn)

            log.info("Subscribing %s to %s" % (func.name, topic_name))

            # Add permission to lambda for sns invocation.
            try:
                lambda_client.add_permission(
                    FunctionName=func.name,
                    StatementId='sns-topic-' + topic_name,
                    SourceArn=arn,
                    Action='lambda:InvokeFunction',
                    Principal='sns.amazonaws.com')
                log.debug("Added permission for sns to invoke lambda")
                # iam eventual consistency and propagation
                time.sleep(self.iam_delay)
            except ClientError as e:
                if e.response['Error']['Code'] != 'ResourceConflictException':
                    raise

            # Subscribe the lambda to the topic.
            topic = self.session.resource('sns').Topic(arn)
            topic.subscribe(Protocol='lambda', Endpoint=func.arn)  # idempotent

    def remove(self, func):
        lambda_client = self.session.client('lambda')
        for topic_arn in self.topic_arns:
            region, topic_name, statement_id = self._parse_arn(topic_arn)

            try:
                response = lambda_client.remove_permission(
                    FunctionName=func.name,
                    StatementId=statement_id)
                log.debug("Removed lambda permission result: %s" % response)
            except ClientError as e:
                if e.response['Error']['Code'] != 'ResourceNotFoundException':
                    raise

            paginator = self.client.get_paginator('list_subscriptions_by_topic')

            class Done(Exception):
                pass
            try:
                for page in paginator.paginate(TopicArn=topic_arn):
                    for subscription in page['Subscriptions']:
                        if subscription['Endpoint'] != func.arn:
                            continue
                        try:
                            response = self.client.unsubscribe(
                                SubscriptionArn=subscription['SubscriptionArn'])
                            log.debug("Unsubscribed %s from %s" %
                                (func.name, topic_name))
                        except ClientError as e:
                            code = e.response['Error']['Code']
                            if code != 'ResourceNotFoundException':
                                raise
                        raise Done  # break out of both for loops
            except Done:
                pass


class BucketSNSNotification(SNSSubscription):
    """ Subscribe a lambda to bucket notifications via SNS. """

    def __init__(self, session_factory, bucket, topic=None):
        # NB: We are overwriting __init__ vs. extending.
        self.session_factory = session_factory
        self.session = session_factory()
        self.topic_arns = self.get_topic(bucket) if topic is None else [topic]
        self.client = self.session.client('sns')

    def get_topic(self, bucket):
        session = local_session(self.session_factory)
        sns = session.client('sns')
        s3 = session.client('s3')

        notifies = bucket['Notification']
        if 'TopicConfigurations' not in notifies:
            notifies['TopicConfigurations'] = []
        all_topics = notifies['TopicConfigurations']
        topic_arns = [t['TopicArn'] for t in all_topics
                      if 's3:ObjectCreated:*' in t['Events']]
        if not topic_arns:
            # No suitable existing topic. Create one.
            topic_arn = sns.create_topic(Name=bucket['Name'])['TopicArn']
            policy = {
                'Statement': [{
                    'Action': 'SNS:Publish',
                    'Effect': 'Allow',
                    'Resource': topic_arn,
                    'Principal': {'Service': 's3.amazonaws.com'}}]}
            sns.set_topic_attributes(
                TopicArn=topic_arn,
                AttributeName='Policy',
                AttributeValue=json.dumps(policy))
            notifies['TopicConfigurations'].append({
                'TopicArn': topic_arn,
                'Events': ['s3:ObjectCreated:*']})
            s3.put_bucket_notification_configuration(Bucket=bucket['Name'],
                NotificationConfiguration=notifies)
            topic_arns = [topic_arn]
        return topic_arns


class ConfigRule(object):
    """Use a lambda as a custom config rule.

    """

    def __init__(self, data, session_factory):
        self.data = data
        self.session_factory = session_factory
        self.session = session_factory()
        self.client = self.session.client('config')

    def __repr__(self):
        return "<ConfigRule>"

    def get_rule_params(self, func):
        # config does not support versions/aliases on lambda funcs
        func_arn = func.arn
        if func_arn.count(':') > 6:
            func_arn, version = func_arn.rsplit(':', 1)

        params = dict(
            ConfigRuleName=func.name,
            Description=func.description,
            Source={
                'Owner': 'CUSTOM_LAMBDA',
                'SourceIdentifier': func_arn,
                'SourceDetails': [{
                    'EventSource': 'aws.config',
                    'MessageType': 'ConfigurationItemChangeNotification'}]
            }
        )

        if isinstance(func, PolicyLambda):
            manager = func.policy.get_resource_manager()
            config_type = manager.get_model().config_type
            params['Scope'] = {
                'ComplianceResourceTypes': [config_type]}
        else:
            params['Scope']['ComplianceResourceTypes'] = self.data.get(
                'resource-types', ())
        return params

    def get(self, rule_name):
        rules = resource_exists(
            self.client.describe_config_rules,
            ConfigRuleNames=[rule_name],
            NotFound="NoSuchConfigRuleException")
        if not rules:
            return rules
        return rules['ConfigRules'][0]

    @staticmethod
    def delta(rule, params):
        # doesn't seem like we have anything mutable at the moment,
        # since we restrict params, maybe reusing the same policy name
        # with a different resource type.
        if rule['Scope'] != params['Scope']:
            return True
        if rule['Source'] != params['Source']:
            return True
        if rule.get('Description', '') != rule.get('Description', ''):
            return True
        return False

    def add(self, func):
        rule = self.get(func.name)
        params = self.get_rule_params(func)

        if rule and self.delta(rule, params):
            log.debug("Updating config rule for %s" % self)
            rule.update(params)
            return self.client.put_config_rule(ConfigRule=rule)
        elif rule:
            log.debug("Config rule up to date")
            return
        try:
            self.session.client('lambda').add_permission(
                FunctionName=func.name,
                StatementId=func.name,
                SourceAccount=func.arn.split(':')[4],
                Action='lambda:InvokeFunction',
                Principal='config.amazonaws.com')
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceConflictException':
                raise

        log.debug("Adding config rule for %s" % func.name)
        return self.client.put_config_rule(ConfigRule=params)

    def remove(self, func):
        rule = self.get(func.name)
        if not rule:
            return
        try:
            self.client.delete_config_rule(
                ConfigRuleName=func.name)
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceNotFoundException':
                raise



# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from common import BaseTest
from c7n.executor import MainThreadExecutor
from c7n.resources.elb import ELB
from c7n.filters import FilterValidationError


class ELBTagTest(BaseTest):

    def test_elb_tag_and_remove(self):
        self.patch(ELB, 'executor_factory', MainThreadExecutor)
        session_factory = self.replay_flight_data('test_elb_tag_and_remove')
        client = session_factory().client('elb')

        policy = self.load_policy({
            'name': 'elb-tag',
            'resource': 'elb',
            'filters': [
                {'LoadBalancerName': 'CloudCustodian'}],
            'actions': [
                {'type': 'tag', 'key': 'xyz', 'value': 'abdef'}]
            },
            session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources), 1)
        tags = client.describe_tags(
            LoadBalancerNames=['CloudCustodian'])['TagDescriptions'][0]['Tags']
        tag_map = {t['Key']: t['Value'] for t in tags}
        self.assertTrue('xyz' in tag_map)

        policy = self.load_policy({
            'name': 'elb-tag',
            'resource': 'elb',
            'filters': [
                {'LoadBalancerName': 'CloudCustodian'}],
            'actions': [
                {'type': 'remove-tag', 'tags': ['xyz']}]
            },
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        tags = client.describe_tags(
            LoadBalancerNames=['CloudCustodian'])['TagDescriptions'][0]['Tags']
        tag_map = {t['Key']: t['Value'] for t in tags}
        self.assertFalse('xyz' in tag_map)

    def test_elb_tags(self):
        self.patch(ELB, 'executor_factory', MainThreadExecutor)
        session_factory = self.replay_flight_data(
            'test_elb_tags')
        policy = self.load_policy({
            'name': 'elb-mark',
            'resource': 'elb',
            'filters': [{"tag:Platform": "ubuntu"}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)

    def test_mark_and_match(self):
        session_factory = self.replay_flight_data(
            'test_elb_mark_and_match')
        policy = self.load_policy({
            'name': 'elb-mark',
            'resource': 'elb',
            'filters': [{"LoadBalancerName": 'CloudCustodian'}],
            'actions': [{
                'type': 'mark-for-op', 'op': 'delete',
                'tag': 'custodian_next', 'days': 1}]},
            session_factory=session_factory)
        resources = policy.run()

        self.assertEqual(len(resources), 1)
        tags = session_factory().client('elb').describe_tags(
            LoadBalancerNames=['CloudCustodian'])['TagDescriptions'][0]['Tags']
        tag_map = {t['Key']: t['Value'] for t in tags}
        self.assertTrue('custodian_next' in tag_map)

        policy = self.load_policy({
            'name': 'elb-mark-filter',
            'resource': 'elb',
            'filters': [
                {'type': 'marked-for-op', 'tag': 'custodian_next',
                 'op': 'delete'}]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)


class ELBInstance(BaseTest):

    def test_instance_filter(self):
        session_factory = self.replay_flight_data(
            'test_elb_instance_filter')
        policy = self.load_policy({
            'name': 'elb-instance',
            'resource': 'elb',
            'filters': [
                {'type': 'instance',
                 'key': 'ImageId',
                 'value': 'ami-40d28157'}],
            }, session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]['LoadBalancerName'], 'balanced')


class HealthCheckProtocolMismatchTest(BaseTest):

    def test_healthcheck_protocol_mismatch(self):
        session_factory = self.replay_flight_data(
            'test_healthcheck_protocol_mismatch')
        policy = self.load_policy({
            'name': 'healthcheck-protocol-mismatch',
            'resource': 'elb',
            'filters': [
                {'type': 'healthcheck-protocol-mismatch'}
            ]},
            session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources), 3)

        # make sure we matched the right load balcners
        elb_names = set([elb['LoadBalancerName'] for elb in resources])
        self.assertEqual(
            elb_names, set(
                ['test-elb-no-listeners',
                 'test-elb-protocol-matches',
                 'test-elb-multiple-listeners']))


class SSLPolicyTest(BaseTest):

    def test_ssl_ciphers(self):
        session_factory = self.replay_flight_data(
            'test_ssl_ciphers')
        policy = self.load_policy({
            'name': 'test-ssl-ciphers',
            'resource': 'elb',
            'filters': [
                {'type': 'ssl-policy',
                 'blacklist': ['Protocol-SSLv2']}
            ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(
            resources[0]['LoadBalancerName'],
            'test-elb-invalid-policy')

    def test_set_ssl_listener_policy(self):
        session_factory = self.replay_flight_data(
            'test_set_ssl_listener')
        client = session_factory().client('elb')
        policy = self.load_policy({
            'name': 'test-set-ssl-listerner',
            'resource': 'elb',
            'filters': [
                {'type': 'ssl-policy',
                 'whitelist': ['AES128-SHA256','Protocol-TLSv1']},
                {
                   'type': 'value',
                   'key': 'LoadBalancerName',
                   'value': 'test-elb',
                   'op': 'eq' }],
            'actions': [
                {'type': 'set-ssl-listener-policy',
                 'name': 'testpolicy',
                 'attributes': ['AES128-SHA256','Protocol-TLSv1']}
            ]},
            session_factory=session_factory)
        resources = policy.run()
        response_pol = client.describe_load_balancers(
            LoadBalancerNames=[
                'test-elb'
            ]
        )
        response_ciphers = client.describe_load_balancer_policies(
            LoadBalancerName='test-elb',
            PolicyNames=['testpolicy-1493768308000']
        )
        curr_pol = [t.encode('UTF8') for t in response_pol[
            'LoadBalancerDescriptions'][0]['ListenerDescriptions'][0]['PolicyNames']]

        curr_ciphers = []
        for x in response_ciphers['PolicyDescriptions'][0]['PolicyAttributeDescriptions']:
            curr_ciphers.append({str(k): str(v) for k, v in x.items()})
        active_ciphers = [x['AttributeName'] for x in curr_ciphers if x['AttributeValue'] == 'true']
        self.assertEqual(
            curr_pol,
            ['AWSConsole-LBCookieStickinessPolicy-test-elb-1493748038333',
             'testpolicy-1493768308000'])
        self.assertEqual(
            active_ciphers,
            ['Protocol-TLSv1', 'AES128-SHA256'])

    def test_ssl_matching(self):
        session_factory = self.replay_flight_data(
            'test_ssl_ciphers')
        policy = self.load_policy({
            'name': 'test-ssl-matching',
            'resource': 'elb',
            'filters': [
                {'type': 'ssl-policy',
                 'matching': '^Protocol-',
                 'whitelist': ['Protocol-TLSv1', 'Protocol-TLSv1.1', 'Protocol-TLSv1.2']}
            ]},
            session_factory=session_factory)
        resources = policy.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(
            resources[0]['LoadBalancerName'],
            'test-elb-invalid-policy')

    def test_filter_validation_no_blacklist(self):
        self.assertRaises(FilterValidationError,
            self.load_policy, {
                'name': 'test-ssl-ciphers',
                'resource': 'elb',
                'filters': [
                    {'type': 'ssl-policy'}
                ]},
                session_factory=None, validate=False)

    def test_filter_validation_blacklist_not_iterable(self):
        self.assertRaises(FilterValidationError,
            self.load_policy, {
                'name': 'test-ssl-ciphers',
                'resource': 'elb',
                'filters': [
                    {'type': 'ssl-policy', 'blacklist': 'single-value'}
                ]},
                session_factory=None, validate=False)


class TestDefaultVpc(BaseTest):

    def test_elb_default_vpc(self):
        session_factory = self.replay_flight_data('test_elb_default_vpc')
        p = self.load_policy(
            {'name': 'elb-default-filters',
             'resource': 'elb',
             'filters': [
                 {'type': 'default-vpc'}]},
            config={'region': 'us-west-2'},
            session_factory=session_factory)

        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(
            resources[0]['LoadBalancerName'], 'test-load-balancer')


class TestModifyVpcSecurityGroupsAction(BaseTest):

    def test_elb_remove_security_groups(self):
        # Test conditions:
        #   - running ELB in default VPC
        #   - security group named TEST-PROD-ONLY-SG exists in VPC and is
        #     attached to test ELB
        session_factory = self.replay_flight_data(
            'test_elb_remove_security_groups')
        client = session_factory().client('ec2')
        default_sg_id = client.describe_security_groups(
            GroupNames=[
                'default',
            ]
        )['SecurityGroups'][0]['GroupId']
        p = self.load_policy(
            {'name': 'elb-modify-security-groups-filter',
             'resource': 'elb',
             'filters': [
                 {'type': 'security-group', 'key': 'GroupName',
                  'value': '(.*PROD-ONLY.*)', 'op': 'regex'}],
             'actions': [
                 {'type': 'modify-security-groups', 'remove': 'matched',
                  'isolation-group': default_sg_id}]
             },
            session_factory=session_factory)

        resources = p.run()
        clean_resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(
            resources[0]['LoadBalancerName'], 'test-load-balancer')
        self.assertEqual(len(clean_resources), 0)

    def test_elb_add_security_group(self):
        # Test conditions:
        #   - running one ELB with 'default' VPC security group attached
        #   - security group named TEST-PROD-ONLY-SG exists in VPC and is not
        #     attached to ELB
        session_factory = self.replay_flight_data(
            'test_elb_add_security_group')

        policy = self.load_policy({
            'name': 'add-sg-to-prod-elb',
            'resource': 'elb',
            'filters': [
                {'type': 'security-group', 'key': 'GroupName',
                 'value': 'default'},
                {'type': 'value', 'key': 'LoadBalancerName',
                 'value': 'test-load-balancer'}],
            'actions': [
                {'type': 'modify-security-groups', 'add': 'sg-411b413c'}
            ]
        },
        session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources[0]['SecurityGroups']), 1)
        after_resources = policy.run()
        self.assertEqual(len(after_resources[0]['SecurityGroups']), 2)

    def test_elb_add_security_groups(self):
        # Test conditions:
        #   - running one ELB with 'default' VPC security group attached
        #   - security groups named TEST-PROD-ONLY-SG, TEST-SG1, and TEST-SG2
        #     exist in VPC - not attached to ELB

        session_factory = self.replay_flight_data(
            'test_elb_add_security_groups')
        policy = self.load_policy({
            'name': 'add-sgs-to-prod-elb',
            'resource': 'elb',
            'filters': [
                {'type': 'security-group', 'key': 'GroupName',
                 'value': 'default'},
                {'type': 'value', 'key': 'LoadBalancerName',
                 'value': 'test-load-balancer'}],
            'actions': [
                {'type': 'modify-security-groups', 'add': [
                    'sg-411b413c', 'sg-8a4b64f7', 'sg-5d4a6520']}
            ]},
            session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources[0]['SecurityGroups']), 1)
        after_resources = policy.run()
        self.assertEqual(len(after_resources[0]['SecurityGroups']), 4)


    def test_elb_remove_all_security_groups(self):
        # Test conditions:
        #   - running one ELB with 'default' and 'TEST-PROD-ONLY-SG' VPC
        #     security groups attached
        session_factory = self.replay_flight_data(
            'test_elb_remove_all_security_groups')
        client = session_factory().client('ec2')

        default_sg_id = client.describe_security_groups(
            GroupNames=[
                'default',
            ]
        )['SecurityGroups'][0]['GroupId']

        policy = self.load_policy({
            'name': 'add-sg-to-prod-elb',
            'resource': 'elb',
            'filters': [
                {'type': 'value', 'key': 'LoadBalancerName',
                 'value': 'test-load-balancer'}],
            'actions': [
                {'type': 'modify-security-groups', 'remove': 'all',
                 'isolation-group': default_sg_id},
            ]},
            session_factory=session_factory)

        resources = policy.run()
        self.assertEqual(len(resources[0]['SecurityGroups']), 2)
        after_resources = policy.run()
        self.assertEqual(len(after_resources[0]['SecurityGroups']), 1)
        # Check that it is indeed the isolation group on the ELB
        self.assertEqual(
            after_resources[0]['SecurityGroups'][0], default_sg_id)
